# This file is part of pycloudlib. See LICENSE file for license information.
"""EC2 Util Functions."""

import base64

import boto3
import botocore

from pycloudlib.util import get_timestamped_tag


def _tag_resource(resource, tag_value=None):
    """Tag a resource with the specified tag.

    This makes finding and deleting resources specific to this testing
    much easier to find.

    Args:
        resource: resource to tag
        tag_value: string, what to tag the item with
    """
    if not tag_value:
        tag_value = get_timestamped_tag(tag="")

    tag = {"Key": "Name", "Value": tag_value}
    resource.create_tags(Tags=[tag])


def _decode_console_output_as_bytes(parsed, **kwargs):
    """Provide console output as bytes in OutputBytes.

    For this to be useful, the session has to have had the
    decode_console_output handler unregistered already.

    https://github.com/boto/botocore/issues/1351

    Args:
        parsed: the raw console output
    """
    if "Output" not in parsed:
        return
    orig = parsed["Output"]
    botocore.handlers.decode_console_output(parsed, **kwargs)
    parsed["OutputBytes"] = base64.b64decode(orig)


def _get_session(access_key_id, secret_access_key, region):
    """Get EC2 session.

    Args:
        access_key_id: user's access key ID
        secret_access_key: user's secret access key
        region: region to login to

    Returns:
        boto3 session object

    """
    mysess = botocore.session.get_session()
    mysess.unregister(
        "after-call.ec2.GetConsoleOutput",
        botocore.handlers.decode_console_output,
    )
    mysess.register(
        "after-call.ec2.GetConsoleOutput", _decode_console_output_as_bytes
    )
    return boto3.Session(
        botocore_session=mysess,
        aws_access_key_id=access_key_id,
        aws_secret_access_key=secret_access_key,
        region_name=region,
    )



# This file is part of pycloudlib. See LICENSE file for license information.
"""LXD instance."""
import json
import re
import time

from pycloudlib.errors import PycloudlibTimeoutError
from pycloudlib.instance import BaseInstance
from pycloudlib.util import subp

MISSING_AGENT_MSG = (
    "Many Xenial images do not support `%s` due to missing lxd-agent:"
    " you may see unavoidable failures.\n"
    "See https://github.com/canonical/pycloudlib/issues/132 for details."
)


class LXDInstance(BaseInstance):
    """LXD backed instance."""

    _type = "lxd"
    _is_vm = None
    _is_ephemeral = None

    def __init__(
        self,
        name,
        key_pair=None,
        execute_via_ssh=True,
        series=None,
        ephemeral=None,
    ):
        """Set up instance.

        Args:
            name: name of instance
            key_pair: SSH key object
            execute_via_ssh: Boolean True to use ssh instead of lxc exec for
                all operations.
            series: Ubuntu release name: xenial, bionic etc.
            ephemeral: Boolean True if instance is ephemeral. If left
                unspecified, ephemeral type will be determined and cached by
                the ephemeral method.
        """
        super().__init__(key_pair=key_pair)

        self._name = name
        self.execute_via_ssh = execute_via_ssh
        self.series = series
        self._is_ephemeral = ephemeral

    def __repr__(self):
        """Create string representation for class."""
        return "LXDInstance(name={})".format(self.name)

    def _run_command(self, command, stdin):
        """Run command in the instance."""
        if self.execute_via_ssh:
            return super()._run_command(command, stdin)

        base_cmd = [
            "lxc",
            "exec",
            self.name,
            "--",
            "sudo",
            "-u",
            self.username,
            "--",
        ]
        return subp(base_cmd + list(command), rcs=None)

    def parse_ip(self, query: dict):
        """Return ip address from lxd query.

        Returns None if no address found
        """
        network = query.get("state", {}).get("network")
        if network is None:
            network = {}
        for nic_cfg in sorted(
            network.values(),
            # prefer nics with host_name defined
            key=lambda cfg: not cfg.get("host_name"),
        ):
            for addr in nic_cfg["addresses"]:
                if addr.get("scope") != "global":
                    continue
                if addr.get("family") == "inet":
                    return addr.get("address")
        self._log.debug(
            "Unable to find valid IP. Found network: %s",
            network,
        )
        return None

    @property
    def is_vm(self):
        """Return boolean if vm type or not.

        Will return False if unknown.

        Returns:
            boolean if virtual-machine
        """
        if self._is_vm is None:
            result = subp(["lxc", "info", self.name])

            try:
                info_type = re.findall(r"Type: (.*)", result)[0]
            except IndexError:
                return False

            self._is_vm = bool(info_type == "virtual-machine")

        return self._is_vm

    @property
    def name(self):
        """Return instance name."""
        return self._name

    @property
    def ip(self):
        """Return IP address of instance.

        Returns:
            IP address assigned to instance.

        Raises: PycloudlibTimeoutError when exhausting retries trying to parse
            lxc list for ip addresses.
        """
        retries = 150

        while retries != 0:
            command = [
                "lxc",
                "query",
                f"/1.0/instances/{self.name}?recursion=1",
            ]
            result = subp(command)
            if result.ok and result.stdout:
                try:
                    info = json.loads(result.stdout)
                except ValueError:
                    self._log.debug(
                        "Unable to parse output of cmd: %s. Expected JSON,"
                        " got: %s. Retrying %d time(s)...",
                        command,
                        result.stdout,
                        retries,
                    )
                else:
                    ip = self.parse_ip(info)
                    if ip:
                        return ip
            retries -= 1
            time.sleep(1)
        raise PycloudlibTimeoutError(
            "Unable to determine IP address after 150 retries."
            " exit:{} stdout: {} stderr: {}".format(
                result.return_code, result.stdout, result.stderr
            )
        )

    @property
    def ephemeral(self):
        """Return boolean if ephemeral or not.

        Will return False if unknown.

        Returns:
            boolean if ephemeral
        """
        if self._is_ephemeral is None:
            result = subp(["lxc", "info", self.name])

            try:
                info_type = re.findall(r"Type: (.*)", result)[0]
                self._is_ephemeral = bool("ephemeral" in info_type)
            except IndexError:
                self._log.debug(
                    "Unable to parse lxc show %s to determine ephemeral type."
                    " Assuming not ephemeral.",
                    self.name,
                )
                self._is_ephemeral = False
        return self._is_ephemeral

    @property
    def state(self):
        """Return current status of instance.

        If unable to get status will return 'Unknown'.

        Returns:
            Reported status from lxc info

        """
        result = subp(["lxc", "info", self.name])
        try:
            return re.findall(r"Status: (.*)", result)[0]
        except IndexError:
            return "Unknown"

    def console_log(self):
        """Return console log.

        Uses the '--show-log' option of console to get the console log
        from an instance.

        Returns:
            bytes of this instance's console

        """
        self._log.debug("getting console log for %s", self.name)
        try:
            return subp(["lxc", "console", self.name, "--show-log"])
        except RuntimeError as exc:
            if "Instance is not container type" not in str(exc):
                raise
            # "Instance is not container type" means we don't support console
            # log for this instance: raise NotImplementedError
            raise NotImplementedError from exc

    def delete(self, wait=True):
        """Delete the current instance.

        By default this will use the '--force' option to prevent the
        need to always stop the instance first. This makes it easier
        to work with ephemeral instances as well, which are deleted
        on stop.

        Args:
            wait: wait for delete
        """
        self._log.debug("deleting %s", self.name)

        subp(["lxc", "delete", self.name, "--force"])

        if wait:
            self.wait_for_delete()

    def delete_snapshot(self, snapshot_name):
        """Delete a snapshot of the instance.

        Args:
            snapshot_name: the name to delete
        """
        self._log.debug("deleting snapshot %s/%s", self.name, snapshot_name)
        subp(["lxc", "delete", "%s/%s" % (self.name, snapshot_name)])

    def edit(self, key, value):
        """Edit the config of the instance.

        Args:
            key: The config key to edit
            value: The new value to set the key to
        """
        self._log.debug("editing %s with %s=%s", self.name, key, value)
        subp(["lxc", "config", "set", self.name, key, value])

    def pull_file(self, remote_path, local_path):
        """Pull file from an instance.

        The remote path must be absolute path with LXD due to the way
        files are pulled off. Specifically, the format is 'name/path'
        with path assumed to start from '/'.

        Args:
            remote_path: path to remote file to pull down
            local_path: local path to put the file
        """
        self._log.debug("pulling file %s to %s", remote_path, local_path)

        if self.execute_via_ssh:
            super().pull_file(remote_path, local_path)
            return

        if self.series == "xenial":
            self._log.warning(MISSING_AGENT_MSG, "lxc file pull")

        if remote_path[0] != "/":
            remote_pwd = self.execute("pwd")
            remote_path = remote_pwd + "/" + remote_path
            self._log.debug("Absolute remote path: %s", remote_path)

        subp(
            [
                "lxc",
                "file",
                "pull",
                "%s%s" % (self.name, remote_path),
                local_path,
            ]
        )

    def push_file(self, local_path, remote_path):
        """Push file to an instance.

        The remote path must be absolute path with LXD due to the way
        files are pulled off. Specifically, the format is 'name/path'
        with path assumed to start from '/'.

        Args:
            local_path: local path to file to push up
            remote_path: path to push file
        """
        self._log.debug("pushing file %s to %s", local_path, remote_path)

        if self.execute_via_ssh:
            super().push_file(local_path, remote_path)
            return

        if self.series == "xenial":
            self._log.warning(MISSING_AGENT_MSG, "lxc file push")

        if remote_path[0] != "/":
            remote_pwd = self.execute("pwd")
            remote_path = remote_pwd + "/" + remote_path
            self._log.debug("Absolute remote path: %s", remote_path)

        subp(
            [
                "lxc",
                "file",
                "push",
                local_path,
                "%s%s" % (self.name, remote_path),
            ]
        )

    def _do_restart(self, force=False, **kwargs):
        """Restart an instance.

        Args:
            force: boolean, force instance to shutdown before restart
        """
        self._log.debug("restarting %s", self.name)

        # Note: even if slightly faster in some cases, do not replace
        # `lxc restart` with stop + start, as ephemeral instances do
        # not survive being stopped, while they do survive restarts.
        cmd = ["lxc", "restart", self.name]
        if force:
            cmd.append("--force")
        subp(cmd)

    def restore(self, snapshot_name):
        """Restore instance from a specific snapshot.

        Args:
            snapshot_name: Name of snapshot to restore from
        """
        self._log.debug(
            "restoring %s from snapshot %s", self.name, snapshot_name
        )
        subp(["lxc", "restore", self.name, snapshot_name])

    def shutdown(self, wait=True, force=False, **kwargs):
        """Shutdown instance.

        Args:
            wait: boolean, wait for instance to shutdown
            force: boolean, force instance to shutdown
        """
        if self.state == "Stopped":
            return

        self._log.debug("shutting down %s", self.name)
        cmd = ["lxc", "stop", self.name]

        if force:
            cmd.append("--force")

        subp(cmd)

        if wait:
            self.wait_for_stop()

    def local_snapshot(self, snapshot_name, stateful=False):
        """Create an LXD snapshot (not a launchable image).

        Args:
            snapshot_name: name to call snapshot
            stateful: boolean, stateful snapshot or not
        """
        self.clean()
        self.shutdown()

        if snapshot_name is None:
            snapshot_name = "{}-snapshot".format(self.name)
        cmd = ["lxc", "snapshot", self.name, snapshot_name]
        if stateful:
            cmd.append("--stateful")

        self._log.debug("creating snapshot %s", snapshot_name)
        subp(cmd)
        return snapshot_name

    def snapshot(self, snapshot_name):
        """Create an image snapshot.

        Snapshot is a bit of a misnomer here. Since "snapshot" in the
        context of clouds means "create a launchable container from
        this instance", we actually need to do a publish here. If you
        need the lxd "snapshot" functionality, use local_snapshot

        Args:
            snapshot_name: name to call snapshot
        """
        self.clean()
        if not self.ephemeral:
            self.shutdown()
        if snapshot_name is None:
            snapshot_name = "{}-snapshot".format(self.name)
        cmd = [
            "lxc",
            "publish",
            "--force",
            self.name,
            "--alias",
            snapshot_name,
        ]

        self._log.debug("Publishing snapshot %s", snapshot_name)
        subp(cmd)
        return "local:{}".format(snapshot_name)

    def start(self, wait=True):
        """Start instance.

        Args:
            wait: boolean, wait for instance to fully start
        """
        if self.state == "Running":
            return

        self._log.debug("starting %s", self.name)
        subp(["lxc", "start", self.name])

        if wait:
            self.wait()

    def wait_for_delete(self):
        """Wait for delete.

        Not used for LXD.
        """

    def wait_for_state(self, desired_state: str, num_retries: int = 100):
        """Wait for instance to reach desired state value.

        :param desired_state: String representing one of lxc instance states
            seen by `lxc ls -s`. For example, ACTIVE, FROZEN, RUNNING, STOPPED
        :param retries: Integer for number of retry attempts before raising a
            PycloudlibTimeoutError.
        """
        self._log.debug("waiting for %s: %s", desired_state, self.name)
        for _ in range(num_retries):
            result = subp(
                [
                    "lxc",
                    "list",
                    "^{}$".format(self.name),
                    "-cs",
                    "--format",
                    "csv",
                ]
            )

            if result == desired_state:
                return
            time.sleep(1)
        raise PycloudlibTimeoutError

    def wait_for_stop(self):
        """Wait for cloud instance to transition to stop state."""
        # Ephemeral instances will not go to STOPPED. They get destroyed.
        if not self.ephemeral:
            self.wait_for_state("STOPPED")

    def _wait_for_instance_start(self):
        """Wait for the cloud instance to be up.

        LXD VMs need to install systemd units upon initialization. There is
        no easy way to do this and also enable them on boot, so an LXD VM
        will reboot as part of its initialization process. It is possible
        we have connected the VM before this reboot occurs, so then any
        following SSH connections will fail.

        The VM doesn't accurately report the number of processes until
        the initialization is fully complete, so block until our number
        of processes isn't -1.
        """
        processes = -1
        for _ in range(600):
            try:
                processes = int(
                    subp(
                        "lxc list --columns N {} --format csv".format(
                            self.name
                        ).split()
                    )
                )
            except ValueError:
                pass
            if processes > -1:
                return
            time.sleep(1)
        raise PycloudlibTimeoutError


class LXDVirtualMachineInstance(LXDInstance):
    """LXD Virtual Machine backed instance."""

    def _run_command(self, command, stdin):
        """Run command in the instance."""
        if self.execute_via_ssh:
            return super()._run_command(command, stdin)

        if self.series == "xenial":
            self._log.warning(MISSING_AGENT_MSG, "lxc exec")

        return super()._run_command(command, stdin)

    def _wait_for_instance_start(self):
        """Wait for the cloud instance to be up.

        LXD VMs need to install systemd units upon initialization. There is
        no easy way to do this and also enable them on boot, so an LXD VM
        will reboot as part of its initialization process. It is possible
        we have connected the VM before this reboot occurs, so then any
        following SSH connections will fail.

        The VM doesn't accurately report the number of processes until
        the initialization is fully complete, so block until our number
        of processes isn't -1.
        """
        # On xenial, we don't install the LXD agent, so we cannot count the
        # number of processes running on the VM. For xenial, we will rely
        # on the other wait methods that are run to guarantee that the instance
        # is running
        if self.series != "xenial":
            super()._wait_for_instance_start()
        else:
            self.wait_for_state(desired_state="RUNNING", num_retries=200)



# This file is part of pycloudlib. See LICENSE file for license information.
"""IBM Cloud type."""

from typing import List, Optional

from ibm_cloud_sdk_core.authenticators import IAMAuthenticator
from ibm_platform_services import ResourceManagerV2
from ibm_vpc import VpcV1
from ibm_vpc.vpc_v1 import Image, ListImagesEnums

from pycloudlib.cloud import BaseCloud
from pycloudlib.config import ConfigFile
from pycloudlib.ibm._util import get_first as _get_first
from pycloudlib.ibm._util import iter_resources as _iter_resources
from pycloudlib.ibm._util import wait_until as _wait_until
from pycloudlib.ibm.errors import IBMException
from pycloudlib.ibm.instance import VPC, IBMInstance
from pycloudlib.instance import BaseInstance
from pycloudlib.util import UBUNTU_RELEASE_VERSION_MAP

DEFAULT_RESOURCE_GROUP: str = "Default"


class IBM(BaseCloud):
    """IBM Virtual Private Cloud Class."""

    _type = "ibm"

    def __init__(
        self,
        tag: str,
        timestamp_suffix: bool = True,
        config_file: Optional[ConfigFile] = None,
        *,
        resource_group: Optional[str] = None,
        vpc: Optional[str] = None,
        api_key: Optional[str] = None,
        region: Optional[str] = None,
        zone: Optional[str] = None,
    ):
        """Initialize the connection to IBM VPC.

        Args:
            tag: string used to name and tag resources with
            timestamp_suffix: Append a timestamped suffix to the tag string.
            config_file: path to pycloudlib configuration file
        """
        super().__init__(
            tag,
            timestamp_suffix,
            config_file,
            required_values=[resource_group, api_key, region],
        )

        self._resource_group = (
            resource_group
            or self.config.get("resource_group")
            or DEFAULT_RESOURCE_GROUP
        )
        self._resource_group_id: Optional[str] = None
        self.region = str(region or self.config.get("region")).lower()
        zone = zone or self.config.get("zone") or f"{self.region}-1"
        self.zone = str(zone).lower()

        self._vpc_name = vpc or self.config.get("vpc")
        self._vpc: Optional[VPC] = None

        self._log.debug("logging into IBM")

        api_key = api_key or self.config.get("api_key")
        authenticator = IAMAuthenticator(api_key)

        self._client = VpcV1(authenticator=authenticator)
        self._client.set_service_url(
            f"https://{self.region}.iaas.cloud.ibm.com/v1"
        )

        self._resource_manager_service = ResourceManagerV2(
            authenticator=authenticator
        )

    @property
    def resource_group_id(self) -> str:
        """Resource Group ID used to create new things under."""
        if self._resource_group_id is None:
            self._resource_group_id = self._get_resource_group_id(
                self._resource_group
            )
        if self._resource_group_id is None:
            raise IBMException(
                f"Resource Group not found: {self._resource_group}"
            )
        return self._resource_group_id

    @property
    def vpc(self) -> VPC:
        """Virtual Private Cloud."""
        if self._vpc is not None:
            return self._vpc

        kwargs = {
            "client": self._client,
            "resource_group_id": self.resource_group_id,
            "region": self.region,
            "zone": self.zone,
        }
        if self._vpc_name is not None:
            self._vpc = VPC.from_existing(
                self.key_pair, name=self._vpc_name, **kwargs
            )
        else:
            self._vpc = VPC.from_default(self.key_pair, **kwargs)

        return self._vpc

    def _get_resource_group_id(
        self, name: Optional[str] = None
    ) -> Optional[str]:
        name = name or f"{self.tag}-rg"

        result = self._resource_manager_service.list_resource_groups(
            name=name
        ).get_result()
        if result.get("resources"):
            return result["resources"][0]["id"]

        return None

    def delete_image(self, image_id: str, **kwargs):
        """Delete an image.

        Args:
            image_id: string, id of the image to delete
            **kwargs: dictionary of other arguments to pass to delete_image
        """
        self._client.delete_image(image_id).get_result()

    def released_image(self, release, *, arch: str = "amd64", **kwargs):
        """ID of the latest released image for a particular release.

        Args:
            release: The release to look for

        Returns:
            A single string with the latest released image ID for the
            specified release.

        """
        list_images_kwargs = {
            "start": None,
            "limit": None,
            "resource_group_id": None,
            "name": None,
            "visibility": ListImagesEnums.Visibility.PUBLIC.value,
        }
        version = UBUNTU_RELEASE_VERSION_MAP[release].replace(".", "-")
        os_name = f"ubuntu-{version}-{arch}"

        # Images are sorted by (created_at, id), thus we return the first
        # one matching the criterion.
        image = _get_first(
            self._client.list_images,
            resource_name="images",
            filter_fn=lambda img: img["operating_system"]["name"] == os_name,
            **list_images_kwargs,
        )
        if image is None:
            raise ValueError(f"Image not found: {os_name}")
        return image["id"]

    def daily_image(self, release: str, **kwargs) -> str:
        """ID of the latest daily image for a particular release.

        Args:
            release: The release to look for

        Returns:
            A single string with the latest daily image ID for the
            specified release.

        """
        self._log.info(
            "There are no daily images in IBM Cloud."
            " Using released image instead"
        )
        return self.released_image(release, **kwargs)

    def image_serial(self, image_id):
        """Find the image serial of the latest daily image for a particular release.

        Args:
            image_id: string, Ubuntu image id

        Returns:
            string, serial of latest image

        """
        raise NotImplementedError(
            "IBM Cloud does not contain Ubuntu daily images"
        )

    def get_instance(self, instance_id: str, **kwargs) -> BaseInstance:
        """Get an instance by id.

        Args:
            instance_id:

        Returns:
            An instance object to use to manipulate the instance further.

        """
        return IBMInstance.find_existing(
            self.key_pair,
            client=self._client,
            instance_id=instance_id,
        )

    def get_or_create_vpc(self, name: str) -> VPC:
        """Get a VPC by name or create it if not found."""
        args = (self.key_pair,)
        kwargs = {
            "client": self._client,
            "name": name,
            "resource_group_id": self.resource_group_id,
            "zone": self.zone,
        }
        try:
            return VPC.from_existing(*args, **kwargs)
        except IBMException:
            return VPC.create(*args, **kwargs)

    def _create_floating_ip(self, name: Optional[str] = None) -> dict:
        name = name or f"{self.tag}-fi"
        proto = {
            "name": name,
            "resource_group": {"id": self.resource_group_id},
            "zone": {"name": self.zone},
        }
        floating_ip = self._client.create_floating_ip(proto).get_result()
        self._log.info("Floating ip created: %s", {floating_ip["name"]})
        return floating_ip

    def launch(
        self,
        image_id: str,
        instance_type: str = "bx2-2x8",
        user_data=None,
        wait: bool = True,
        *,
        name: Optional[str] = None,
        vpc: Optional[VPC] = None,
        **kwargs,
    ) -> BaseInstance:
        """Launch an instance.

        Args:
            image_id: string, image ID to use for the instance
            instance_type: string, type of instance to create
            user_data: used by cloud-init to run custom scripts/configuration
            wait: wait for instance to be live
            name: instance name
            vpc: VPC to allocate the instance in. If not given, the instance
            will be allocated in the zone's default VPC.
            **kwargs: dictionary of other arguments to pass to launch

        Returns:
            An instance object to use to manipulate the instance further.

        """
        if not image_id:
            raise ValueError(
                f"{self._type} launch requires image_id param."
                f" Found: {image_id}"
            )

        vpc = vpc or self.vpc
        floating_ip_name = f"{name}-fi" if name else None
        name = name or f"{self.tag}-vm"

        raw_instance = IBMInstance.create_raw_instance(
            client=self._client,
            name=name,
            image_id=image_id,
            vpc=vpc,
            instance_type=instance_type,
            resource_group_id=self.resource_group_id,
            zone=self.zone,
            user_data=user_data,
            key_id=self._get_or_create_key(),
            **kwargs,
        )

        floating_ip = self._create_floating_ip(name=floating_ip_name)
        instance = IBMInstance.with_floating_ip(
            self.key_pair,
            client=self._client,
            instance=raw_instance,
            floating_ip=floating_ip,
        )

        if wait:
            instance.wait()

        return instance

    def snapshot(
        self, instance: IBMInstance, clean: bool = True, **kwargs
    ) -> str:
        """Snapshot an instance and generate an image from it.

        Args:
            instance: Instance to snapshot
            clean: run instance clean method before taking snapshot

        Returns:
            An image id

        """
        if clean:
            instance.clean()

        instance.shutdown()

        self._log.debug("creating snapshot from instance %s", instance.id)

        image_prototype = {
            "name": f"{self.tag}-image",
            "resource_group": {"id": self.resource_group_id},
            "source_volume": {"id": instance.boot_volume_id},
        }

        snapshot_id = self._client.create_image(image_prototype).get_result()[
            "id"
        ]

        timeout_seconds = 300
        _wait_until(
            lambda: self._client.get_image(snapshot_id).get_result()["status"]
            == Image.StatusEnum.AVAILABLE.value,
            timeout_seconds=timeout_seconds,
            timeout_msg_fn=lambda: (
                f"Snapshot not available after {timeout_seconds} seconds. "
                "Check IBM VPC console."
            ),
        )
        return snapshot_id

    def list_keys(self) -> List[str]:
        """List ssh key names present on the cloud for accessing instances.

        Returns:
           A list of strings of key pair names accessible to the cloud.

        """
        return list(
            _iter_resources(
                self._client.list_keys,
                resource_name="keys",
                map_fn=lambda key: key["name"],
            )
        )

    def delete_key(self, name: str):
        """Delete SSH key by name."""
        key = _get_first(
            self._client.list_keys,
            resource_name="keys",
            filter_fn=lambda key: key["name"] == name,
        )
        if not key:
            return
        self._log.debug("Deleting SSH key: %s", name)
        self._client.delete_key(key["id"])

    def _get_or_create_key(self) -> str:
        key = _get_first(
            self._client.list_keys,
            resource_name="keys",
            filter_fn=lambda key: key["name"] == self.key_pair.name,
        )
        if key is not None:
            return key["id"]

        self._log.info("Creating SSH key: %s", self.key_pair.name)
        return self._client.create_key(
            public_key=self.key_pair.public_key_content,
            name=self.key_pair.name,
            resource_group={"id": self.resource_group_id},
        ).get_result()["id"]



# This file is part of pycloudlib. See LICENSE file for license information.
"""Base Key Class."""

import os


class KeyPair:
    """Key Class."""

    def __init__(self, public_key_path, private_key_path=None, name=None):
        """Initialize key class to generate key or reuse existing key.

        The public key path is given then the key is stored and the
        private key is assumed to be named the same minus '.pub'
        otherwise the user should also specify the private key path.

        Args:
            public_key_path: Path to public key
            private_key_path: Path to private key
            name: Name to reference key by in clouds
        """
        self.name = name
        self.public_key_path = public_key_path
        if private_key_path:
            self.private_key_path = private_key_path
        else:
            self.private_key_path = self.public_key_path.replace(".pub", "")

        self.public_key_path = os.path.expanduser(self.public_key_path)
        self.private_key_path = os.path.expanduser(self.private_key_path)

    def __str__(self):
        """Create string representation of class."""
        return "KeyPair({}, {}, name={})".format(
            self.private_key_path, self.public_key_path, self.name
        )

    @property
    def public_key_content(self):
        """Read the contents of the public key.

        Returns:
            output of public key

        """
        return open(self.public_key_path, encoding="utf-8").read()



"""Common GCE utils."""
import os
from urllib.error import HTTPError

import google.auth
from google.oauth2 import service_account

from pycloudlib.gce.errors import GceException


def raise_on_error(response):
    """Look for errors in response and raise if found."""
    if "httpErrorStatusCode" in response:
        raise HTTPError(
            url=response["selfLink"],
            code=response["httpErrorStatusCode"],
            msg=response["httpErrorMessage"],
            hdrs={},
            fp=None,
        )
    if "error" in response:
        raise GceException(
            "Received error(s)!\n"
            "Errors: {}".format(response["error"]["errors"])
        )


def get_credentials(credentials_path):
    """Get GCE account credentials.

    Try service account credentials first. If those fail, try the environment
    """
    credentials_path = os.path.expandvars(os.path.expanduser(credentials_path))
    if credentials_path:
        try:
            return service_account.Credentials.from_service_account_file(
                credentials_path
            )
        except ValueError:
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = credentials_path
    return google.auth.default()[0]



# This file is part of pycloudlib. See LICENSE file for license information.
"""Private utilities for IBM cloud."""
from functools import partial
from time import sleep
from typing import Callable, Iterator, Optional

from ibm_vpc import DetailedResponse

from pycloudlib.errors import PycloudlibTimeoutError
from pycloudlib.util import get_query_param

_IBMCallable = Callable[..., DetailedResponse]


def iter_pages(
    op: _IBMCallable, *, start: Optional[str] = None, **kwargs
) -> Iterator[DetailedResponse]:
    """Lazily iterate over a paginated endpoint."""
    op = partial(op, **kwargs)
    detailed_response: DetailedResponse = op(start=start)
    result = detailed_response.get_result()
    yield detailed_response

    while result.get("next") is not None:
        next_url = result["next"]["href"]
        next_start = get_query_param(next_url, param="start")[0]
        detailed_response = op(start=next_start)
        result = detailed_response.get_result()
        yield detailed_response


def iter_resources(
    op, *, resource_name: str, filter_fn=None, map_fn=None, **kwargs
) -> Iterator:
    """Iterate over the resources, optionally mapped and or filtered."""
    filter_fn = filter_fn or (lambda _: True)
    map_fn = map_fn or (lambda x: x)

    for resp in iter_pages(op, **kwargs):
        resources = resp.get_result().get(resource_name, [])
        yield from map(map_fn, filter(filter_fn, resources))


def get_first(
    op: _IBMCallable,
    *,
    resource_name: str,
    filter_fn: Optional[Callable[..., bool]] = None,
    **kwargs,
) -> Optional[dict]:
    """Get first resource filtered by `filter_fn`."""
    try:
        return next(
            iter_resources(
                op, resource_name=resource_name, filter_fn=filter_fn, **kwargs
            )
        )
    except StopIteration:
        return None


def wait_until(
    check_fn: Callable,
    *args,
    timeout_seconds: int,
    timeout_msg_fn: Callable[..., str],
    raise_on_fail: bool = True,
) -> bool:
    """Wait `timeout_seconds` until `check_fn` evaluates to true."""
    for _ in range(timeout_seconds):
        if check_fn(*args) is True:
            return True
        sleep(1)
    if raise_on_fail:
        raise PycloudlibTimeoutError(timeout_msg_fn())
    return False



"""Tests related to pycloudlib.azure.cloud module."""
from io import StringIO

import mock
import pytest

from pycloudlib.azure.cloud import Azure

CONFIG = """\
[azure]

"""


# Disable this one because we're intentionally testing a protected member
# pylint: disable=protected-access
class TestCreateNetworkInterfaceClient:
    """Tests covering _create_network_interface_client method."""

    @pytest.mark.parametrize(
        "inbound_ports",
        (
            ["3128", "8080"],
            None,
        ),
    )
    @mock.patch("pycloudlib.azure.util.get_client")
    def test_create_network_interface_with_inbound_ports(
        self, m_get_client, inbound_ports
    ):
        """Test method handling of inbound_ports."""
        resource_client_mock = mock.MagicMock()
        resource_group_mock = mock.MagicMock()
        resource_mock = mock.MagicMock()

        network_client_mock = mock.MagicMock()
        network_group_mock = mock.MagicMock()
        compute_client_mock = mock.MagicMock()
        m_get_client.side_effect = [
            resource_client_mock,
            network_client_mock,
            compute_client_mock,
        ]

        type(resource_mock).name = mock.PropertyMock(
            return_value="resource_group"
        )
        type(resource_client_mock).resource_groups = mock.PropertyMock(
            return_value=resource_group_mock
        )
        resource_group_mock.create_or_update.return_value = resource_mock
        type(network_client_mock).network_security_groups = mock.PropertyMock(
            return_value=network_group_mock
        )

        instance = Azure(
            tag="tag",
            timestamp_suffix=False,
            config_file=StringIO(CONFIG),
            region="location",
        )
        instance._create_network_security_group(inbound_ports=inbound_ports)

        expected_security_rules = [
            {
                "name": "port-22",
                "priority": 300,
                "protocol": "TCP",
                "access": "Allow",
                "direction": "Inbound",
                "sourceAddressPrefix": "*",
                "sourcePortRange": "*",
                "destinationAddressPrefix": "*",
                "destinationPortRange": "22",
            },
        ]

        if inbound_ports:
            expected_security_rules.extend(
                [
                    {
                        "name": "port-3128",
                        "priority": 310,
                        "protocol": "TCP",
                        "access": "Allow",
                        "direction": "Inbound",
                        "sourceAddressPrefix": "*",
                        "sourcePortRange": "*",
                        "destinationAddressPrefix": "*",
                        "destinationPortRange": "3128",
                    },
                    {
                        "name": "port-8080",
                        "priority": 320,
                        "protocol": "TCP",
                        "access": "Allow",
                        "direction": "Inbound",
                        "sourceAddressPrefix": "*",
                        "sourcePortRange": "*",
                        "destinationAddressPrefix": "*",
                        "destinationPortRange": "8080",
                    },
                ]
            )

        expected_calls = [
            mock.call(
                resource_group_name="resource_group",
                network_security_group_name="tag-sgn",
                parameters={
                    "location": "location",
                    "security_rules": expected_security_rules,
                },
            ),
        ]

        assert (
            expected_calls
            == network_group_mock.begin_create_or_update.call_args_list
        )



# This file is part of pycloudlib. See LICENSE file for license information.
"""Main pycloud module __init__."""

import logging

from pycloudlib.azure.cloud import Azure
from pycloudlib.ec2.cloud import EC2
from pycloudlib.gce.cloud import GCE
from pycloudlib.ibm.cloud import IBM
from pycloudlib.lxd.cloud import LXD, LXDContainer, LXDVirtualMachine
from pycloudlib.oci.cloud import OCI
from pycloudlib.openstack.cloud import Openstack

__all__ = [
    "Azure",
    "EC2",
    "GCE",
    "IBM",
    "LXD",
    "LXDContainer",
    "LXDVirtualMachine",
    "OCI",
    "Openstack",
]

logging.getLogger(__name__).addHandler(logging.NullHandler())



"""Tests related to pycloudlib.cloud module."""
from io import StringIO
from textwrap import dedent

import mock
import pytest

from pycloudlib.cloud import BaseCloud

# mock module path
MPATH = "pycloudlib.cloud."
CONFIG = """\
[base]

"""


class CloudSubclass(BaseCloud):
    """Create a concrete subclass of BaseCloud for testing."""

    def delete_image(self):
        """Skeletal delete_image."""

    def released_image(self, release, **kwargs):
        """Skeletal released_image."""

    def daily_image(self, release, **kwargs):
        """Skeletal daily_image."""

    def image_serial(self, image_id):
        """Skeletal image_serial."""

    def get_instance(self, instance_id):
        """Skeletal get_instance."""

    def launch(
        self, image_id, instance_type=None, user_data=None, wait=True, **kwargs
    ):
        """Skeletal launch."""

    def snapshot(self, instance, clean=True, **kwargs):
        """Skeletal snapshot."""

    def list_keys(self):
        """Skeletal list_keys."""


class TestBaseCloud:
    """Tests covering BaseCloud intialization."""

    def test_base_cloud_is_abstract(self):
        """The BaseCloud needs a concrete subclass to __init__."""
        with pytest.raises(TypeError) as exc_info:
            BaseCloud(  # pylint: disable=E0110
                tag="", config_file=StringIO(CONFIG)
            )
        assert "Can't instantiate abstract class BaseCloud" in str(
            exc_info.value
        )

    @pytest.mark.parametrize(
        "tag,timestamp_suffix,expected_tag",
        (
            ("a", None, "a-timestamp"),
            ("a", True, "a-timestamp"),
            ("a", False, "a"),
        ),
    )
    @mock.patch(MPATH + "get_timestamped_tag", return_value="a-timestamp")
    @mock.patch(MPATH + "getpass.getuser", return_value="crashoverride")
    def test_init_sets_timestamped_tag(
        self,
        _m_getuser,
        _m_get_timestamped_tag,
        tag,
        timestamp_suffix,
        expected_tag,
    ):
        """The timestamp_suffix param of true adds a tag timestamp suffix."""
        args = {"tag": tag}
        if timestamp_suffix in (True, False):
            args["timestamp_suffix"] = timestamp_suffix
        mycloud = CloudSubclass(config_file=StringIO(CONFIG), **args)
        assert expected_tag == mycloud.tag

    @mock.patch(MPATH + "getpass.getuser", return_value="root")
    def test_init_sets_key_pair_based_on_getuser(self, _m_getuser):
        """
        The default key_pair for the cloud is based on the current user.

        The root user is used as it's guaranteed to exist and has a
        well known $HOME. Also its $HOME is not under /home, so this
        verifies that we're not hardcoding /home/<user> paths.
        """
        mycloud = CloudSubclass(
            tag="tag", timestamp_suffix=False, config_file=StringIO(CONFIG)
        )
        assert mycloud.key_pair.name == "root"
        assert mycloud.key_pair.private_key_path == ("/root/.ssh/id_rsa")
        assert mycloud.key_pair.public_key_path == ("/root/.ssh/id_rsa.pub")

    def test_init_sets_key_pair_from_config(self):
        """The key_pair is set from the config file."""
        mycloud = CloudSubclass(
            tag="tag",
            timestamp_suffix=False,
            config_file=StringIO(
                dedent(
                    """
                [base]

                key_name = "some_name"
                public_key_path = "/home/asdf/.ssh/id_rsa.pub"
                private_key_path = "/home/asdf/.ssh/my_key"
                """
                )
            ),
        )
        assert mycloud.key_pair.name == "some_name"
        assert mycloud.key_pair.public_key_path == "/home/asdf/.ssh/id_rsa.pub"
        assert mycloud.key_pair.private_key_path == "/home/asdf/.ssh/my_key"

    def test_missing_private_key_in_ssh_config(self):
        """The key_pair assumes the private key name."""
        mycloud = CloudSubclass(
            tag="tag",
            timestamp_suffix=False,
            config_file=StringIO(
                dedent(
                    """
                [base]

                key_name = "some_name"
                public_key_path = "/home/asdf/.ssh/id_rsa.pub"
                """
                )
            ),
        )
        assert mycloud.key_pair.name == "some_name"
        assert mycloud.key_pair.public_key_path == "/home/asdf/.ssh/id_rsa.pub"
        assert mycloud.key_pair.private_key_path == "/home/asdf/.ssh/id_rsa"



# This file is part of pycloudlib. See LICENSE file for license information.
"""Module for IBM util tests."""
from unittest import mock

import pytest
from ibm_vpc import DetailedResponse
from mock import MagicMock

from pycloudlib.errors import PycloudlibTimeoutError
from pycloudlib.ibm._util import (
    get_first,
    iter_pages,
    iter_resources,
    wait_until,
)

M_PATH = "pycloudlib.ibm._util."


def build_detailed_response(result) -> mock.MagicMock:
    """Build a `DetailedResponse` mock with `result`."""
    response = mock.MagicMock(spec=DetailedResponse)
    response.get_result.return_value = result
    return response


class TestIterPages:
    """`iter_pages` tests."""

    def test_one_page(self):
        """The ibm operation returns one page."""
        response = build_detailed_response({"next": None})
        m_ibm_operation = mock.MagicMock()
        m_ibm_operation.return_value = response

        pages = iter_pages(m_ibm_operation)

        assert response == next(pages)
        assert [mock.call(start=None)] == m_ibm_operation.call_args_list
        with pytest.raises(StopIteration):
            next(pages)

    def test_two_pages(self):
        """The ibm operation returns two page."""
        response_1 = build_detailed_response(
            {"next": {"href": "http://ibm.com/resouce?start=<start_pointer>"}}
        )
        response_2 = build_detailed_response({"next": None})
        m_op = mock.MagicMock()
        m_op.side_effect = (response_1, response_2)

        pages = iter_pages(m_op)

        assert response_1 == next(pages)
        assert [mock.call(start=None)] == m_op.call_args_list

        assert response_2 == next(pages)
        assert [
            mock.call(start=None),
            mock.call(start="<start_pointer>"),
        ] == m_op.call_args_list

        with pytest.raises(StopIteration):
            next(pages)


class TestGetFirst:
    """`get_first` tests."""

    @mock.patch(M_PATH + "iter_pages")
    def test_one_page_no_filter(self, m_iter_pages):
        """The ibm operation returns one page and no filter is applied."""
        m_ibm_operation = mock.MagicMock()
        responses = (build_detailed_response({"resources": ["x", "y"]}),)
        m_iter_pages.return_value = responses
        kwargs = {"a": "a", "b": "b"}

        assert "x" == get_first(
            m_ibm_operation, resource_name="resources", **kwargs
        )
        assert [
            mock.call(m_ibm_operation, **kwargs)
        ] == m_iter_pages.call_args_list

    @mock.patch(M_PATH + "iter_pages")
    def test_one_page_filter(self, m_iter_pages):
        """The ibm operation returns one page and a filter is applied."""
        m_ibm_operation = mock.MagicMock()
        responses = (build_detailed_response({"resources": ["x", "y"]}),)
        m_iter_pages.return_value = responses
        kwargs = {"a": "a", "b": "b"}

        assert "y" == get_first(
            m_ibm_operation,
            resource_name="resources",
            filter_fn=lambda resource: resource == "y",
            **kwargs,
        )
        assert [
            mock.call(m_ibm_operation, **kwargs)
        ] == m_iter_pages.call_args_list


class TestGetAll:
    """`get_all` tests."""

    @mock.patch(M_PATH + "iter_pages")
    def test_two_pages_no_map(self, m_iter_pages):
        """`iter_pages` returns two page and no mapping is applied."""
        m_ibm_operation = mock.MagicMock()
        responses = (
            build_detailed_response({"resources": ["x", "y"]}),
            build_detailed_response({"resources": ["z"]}),
        )
        m_iter_pages.return_value = responses
        kwargs = {"a": "a", "b": "b"}

        assert ["x", "y", "z"] == list(
            iter_resources(
                m_ibm_operation,
                resource_name="resources",
                **kwargs,
            )
        )
        assert [
            mock.call(m_ibm_operation, **kwargs)
        ] == m_iter_pages.call_args_list

    @mock.patch(M_PATH + "iter_pages")
    def test_two_pages_map(self, m_iter_pages):
        """`iter_pages` returns two page and a mapping is applied."""
        m_ibm_operation = mock.MagicMock()
        responses = (
            build_detailed_response({"resources": ["x", "y"]}),
            build_detailed_response({"resources": ["z"]}),
        )
        m_iter_pages.return_value = responses
        kwargs = {"a": "a", "b": "b"}

        assert ["x^^", "y^^", "z^^"] == list(
            iter_resources(
                m_ibm_operation,
                resource_name="resources",
                map_fn=lambda resource: resource + "^^",
                **kwargs,
            )
        )
        assert [
            mock.call(m_ibm_operation, **kwargs)
        ] == m_iter_pages.call_args_list


class TestWaitUntil:
    """`wait_until` tests."""

    @mock.patch(M_PATH + "sleep")
    def test_eventual_true(self, m_sleep):
        """`check_fn` does succeed after 3 calls."""
        check_fn = MagicMock()
        check_fn.side_effect = (False, False, True)

        assert True is wait_until(
            check_fn=check_fn,
            timeout_seconds=20,
            timeout_msg_fn=lambda: "<msg>",
        )
        assert [mock.call(1), mock.call(1)] == m_sleep.call_args_list

    @mock.patch(M_PATH + "sleep")
    def test_false_raises(self, m_sleep):
        """`check_fn` never succeed and exception is raised."""
        check_fn = MagicMock()
        check_fn.return_value = False

        with pytest.raises(PycloudlibTimeoutError, match="<msg>"):
            wait_until(
                check_fn=check_fn,
                timeout_seconds=20,
                timeout_msg_fn=lambda: "<msg>",
                raise_on_fail=True,
            )
        assert [mock.call(1)] * 20 == m_sleep.call_args_list

    @mock.patch(M_PATH + "sleep")
    def test_false(self, m_sleep):
        """`check_fn` never succeed and exception is never raised."""
        check_fn = MagicMock()
        check_fn.return_value = False

        assert False is wait_until(
            check_fn=check_fn,
            timeout_seconds=20,
            timeout_msg_fn=lambda: "<msg>",
            raise_on_fail=False,
        )
        assert [mock.call(1)] * 20 == m_sleep.call_args_list



"""Module containing errors specific to openstack."""
from pycloudlib.errors import PycloudlibException


class OpenStackError(PycloudlibException):
    """OpenStack exception root."""


class OpenStackFlavorNotFound(OpenStackError):
    """Raised when an OpenStack's flavor is not found."""



# This file is part of pycloudlib. See LICENSE file for license information.
"""Base Result Class."""


class Result(str):  # pylint: disable=too-many-ancestors
    """Result Class."""

    def __init__(self, stdout, stderr="", return_code=0):
        """Initialize class."""
        super().__init__()

        self.stdout = stdout
        self.stderr = stderr
        self.return_code = return_code

    def __new__(cls, stdout, stderr, return_code):
        """Create new class."""
        obj = str.__new__(cls, stdout)
        obj.stderr = stderr
        obj.return_code = return_code
        return obj

    def __bool__(self):
        """Boolean behavior."""
        return self.ok

    @property
    def failed(self):
        """Return boolean if result was failure."""
        return not self.ok

    @property
    def ok(self):
        """Return boolean if result was success."""
        if self.return_code == 0:
            return True
        return False



"""Openstack instance type."""
import time
from itertools import chain

import openstack
from openstack.exceptions import (
    BadRequestException,
    ConflictException,
    ResourceNotFound,
)

from pycloudlib.errors import PycloudlibError
from pycloudlib.instance import BaseInstance


class OpenstackInstance(BaseInstance):
    """Openstack instance object."""

    _type = "openstack"

    def __init__(self, key_pair, instance_id, network_id, connection=None):
        """Set up the instance.

        Args:
            key_pair: A KeyPair for SSH interactions
            instance_id: The instance id representing the cloud instance
            network_id: if of the network this instance was created on
            connection: The connection used to create this instance.
                If None, connection will be created.
        """
        super().__init__(key_pair)

        if not connection:
            connection = openstack.connect()
        self.network_id = network_id
        self.conn = connection

        self.server = self.conn.compute.get_server(instance_id)

        self.delete_floating_ip = False
        self.floating_ip = self._get_existing_floating_ip()
        if self.floating_ip is None:
            self.floating_ip = self._create_and_attach_floating_ip()
            self.delete_floating_ip = True
        self.added_local_ports = []

    def _get_existing_floating_ip(self):
        server_addresses = chain(*self.server.addresses.values())
        server_ips = [addr["addr"] for addr in server_addresses]
        for floating_ip in self.conn.network.ips():
            if floating_ip["floating_ip_address"] in server_ips:
                return floating_ip
        return None

    def _create_and_attach_floating_ip(self):
        floating_ip = self.conn.create_floating_ip(wait=True)
        tries = 30
        for _ in range(tries):
            try:
                self.conn.compute.add_floating_ip_to_server(
                    self.server, floating_ip.floating_ip_address
                )
                break
            except BadRequestException as e:
                if "Instance network is not ready yet" in str(e):
                    time.sleep(1)
                    continue
                raise e
        return floating_ip

    def __repr__(self):
        """Create string representation of class."""
        return "{}(instance_id={})".format(
            self.__class__.__name__, self.server.id
        )

    @property
    def name(self):
        """Return instance name."""
        return self.server.name

    @property
    def ip(self):
        """Return IP address of instance."""
        return self.floating_ip.floating_ip_address

    def console_log(self):
        """Return the instance console log."""
        start = time.time()
        while time.time() < start + 180:
            response = self.conn.compute.get_server_console_output(self.server)
            if response:
                return response["output"]
            self._log.debug("Console output not yet available; sleeping")
            time.sleep(5)
        return "No console output"

    def delete(self, wait=True):
        """Delete the instance.

        Args:
            wait: wait for instance to be deleted
        """
        try:
            self.conn.compute.delete_server(self.server.id)
        finally:
            if self.delete_floating_ip:
                self.conn.delete_floating_ip(self.floating_ip.id)
            for port_id in self.added_local_ports:
                self.conn.network.delete_port(
                    port=port_id, ignore_missing=True
                )
        if wait:
            self.wait_for_delete()

    def _do_restart(self, **kwargs):
        """Restart the instance."""
        self.shutdown(wait=True)
        self.start(wait=False)

    def shutdown(self, wait=True, **kwargs):
        """Shutdown the instance.

        Args:
            wait: wait for the instance to shutdown
        """
        self.conn.compute.stop_server(self.server)
        if wait:
            self.wait_for_stop()

    def start(self, wait=True):
        """Start the instance.

        Args:
            wait: wait for the instance to start.
        """
        try:
            self.conn.compute.start_server(self.server)
        except ConflictException as e:
            # We can get an exception here if the instance is already started
            if "while it is in vm_state active" in str(e):
                return
        if wait:
            self.wait()

    def _wait_for_instance_start(self):
        """Wait for instance to be up."""
        self.conn.compute.wait_for_server(self.server, status="ACTIVE")

    def wait_for_delete(self):
        """Wait for instance to be deleted."""
        try:
            self.conn.compute.wait_for_server(self.server, status="DELETED")
        except ResourceNotFound:
            # We can 404 here is instance is already deleted
            pass

    def wait_for_stop(self):
        """Wait for instance stop."""
        self.conn.compute.wait_for_server(self.server, status="SHUTOFF")

    def add_network_interface(self) -> str:
        """Add nic to running instance.

        Returns IP address in string form
        """
        port = self.conn.network.create_port(
            network_id=self.network_id,
        )
        self.added_local_ports.append(port.id)
        interface = self.conn.compute.create_server_interface(
            server=self.server.id, port_id=port.id
        )
        return interface["fixed_ips"][0]["ip_address"]

    def _get_port_id_by_ip(self, ip_address: str):
        ports = self.conn.network.ports()
        for port in ports:
            for ip in port["fixed_ips"]:
                if ip["ip_address"] == ip_address:
                    return port
        raise PycloudlibError(
            "Could not find port with IP: {}".format(ip_address)
        )

    def remove_network_interface(self, ip_address: str):
        """Remove nic from running instance."""
        port = self._get_port_id_by_ip(ip_address)
        self.conn.network.delete_port(
            port=port.id,
        )
        try:
            self.added_local_ports.remove(port.id)
        except ValueError:
            self._log.warning(
                "Expected port to be in added_local_ports list "
                "but was not: %s",
                port.id,
            )



"""Tests related to pycloudlib.lxd.cloud module."""
import contextlib
import io
from unittest import mock

import pytest

from pycloudlib.lxd.cloud import LXDContainer

M_PATH = "pycloudlib.lxd.cloud."

CONFIG = """\
[lxd]

"""


@contextlib.contextmanager
def does_not_raise():
    """Provide alternate contextmanager to use where no errors are raised."""
    yield


class TestLaunch:
    """Tests covering pycloudlib.lxd.cloud.launch method."""

    @pytest.mark.parametrize(
        "image_id,expectation",
        (
            ("some-img", does_not_raise()),
            ("", pytest.raises(ValueError)),
            (None, pytest.raises(ValueError)),
        ),
    )
    @mock.patch(M_PATH + "_images.find_release")
    def test_launch_validates_image_id(
        self, m_find_release, image_id, expectation
    ):
        """Validate image_id or raise exceptions before calling init/start."""
        m_find_release.return_value = "bionic"
        cloud = LXDContainer(tag="test", config_file=io.StringIO(CONFIG))
        init_kwargs = {
            "image_id": image_id,
            "instance_type": "inst_type",
            "user_data": "ud",
            "wait": False,
            "name": "name",
            "ephemeral": True,
            "network": "netname",
            "storage": "storagename",
            "profile_list": ["profile1"],
            "config_dict": {"user.custom": "val"},
            "execute_via_ssh": True,
        }
        inst = mock.MagicMock()
        with expectation:
            with mock.patch.object(cloud, "init") as lxd_init:
                lxd_init.return_value = inst
                inst = cloud.launch(**init_kwargs)
                wait_val = init_kwargs.pop("wait")
                assert lxd_init.call_args_list == [
                    mock.call(
                        name="name",
                        image_id="some-img",
                        ephemeral=True,
                        network="netname",
                        storage="storagename",
                        inst_type="inst_type",
                        profile_list=["profile1"],
                        user_data="ud",
                        config_dict={"user.custom": "val"},
                        execute_via_ssh=True,
                    )
                ]
                # pylint: disable=no-member
                assert inst.start.call_args_list == [mock.call(wait_val)]
                # pylint: disable=no-member

        if not image_id:
            assert lxd_init.call_count == 0
            assert inst.start.call_count == 0


class TestProfileCreation:
    """Tests covering pycloudlib.lxd.cloud.create_profile method."""

    @mock.patch("pycloudlib.lxd.cloud.subp")
    def test_create_profile_that_already_exists(self, m_subp):
        """Tests creating a profile that already exists."""
        m_subp.return_value = """
            - name: test_profile
        """
        cloud = LXDContainer(tag="test", config_file=io.StringIO(CONFIG))

        fake_stdout = io.StringIO()
        with contextlib.redirect_stdout(fake_stdout):
            cloud.create_profile(
                profile_name="test_profile", profile_config="profile_config"
            )

        expected_msg = "The profile named test_profile already exists"
        assert expected_msg in fake_stdout.getvalue().strip()
        assert m_subp.call_args_list == [
            mock.call(["lxc", "profile", "list", "--format", "yaml"])
        ]

    @mock.patch("pycloudlib.lxd.cloud.subp")
    def test_create_profile_that_already_exists_with_force(self, m_subp):
        """Tests creating an existing profile with force parameter."""
        m_subp.return_value = """
            - name: test_profile
        """
        cloud = LXDContainer(tag="test", config_file=io.StringIO(CONFIG))
        profile_name = "test_profile"
        profile_config = "profile_config"

        cloud.create_profile(
            profile_name=profile_name,
            profile_config=profile_config,
            force=True,
        )

        assert m_subp.call_args_list == [
            mock.call(["lxc", "profile", "list", "--format", "yaml"]),
            mock.call(["lxc", "profile", "delete", profile_name]),
            mock.call(["lxc", "profile", "create", profile_name]),
            mock.call(
                ["lxc", "profile", "edit", profile_name], data=profile_config
            ),
        ]

    @mock.patch("pycloudlib.lxd.cloud.subp")
    def test_create_profile_that_does_not_exist(self, m_subp):
        """Tests creating a new profile."""
        m_subp.return_value = """
            - name: other_profile
        """
        cloud = LXDContainer(tag="test", config_file=io.StringIO(CONFIG))
        profile_name = "other_profile_v1"
        profile_config = "profile_config"

        cloud.create_profile(
            profile_name=profile_name, profile_config=profile_config
        )

        assert m_subp.call_args_list == [
            mock.call(["lxc", "profile", "list", "--format", "yaml"]),
            mock.call(["lxc", "profile", "create", profile_name]),
            mock.call(
                ["lxc", "profile", "edit", profile_name], data=profile_config
            ),
        ]



"""Tests related to pycloudlib.lxd.defaults module."""
import hashlib

import pytest

from pycloudlib.lxd.defaults import LXC_PROFILE_VERSION, base_vm_profiles


class TestLXDProfilesWereNotModified:
    """Test covering if profiles were not accidentally changed."""

    # This dict represents a mapping between the profile version and the
    # md5sum associated with it. Whenever we have a new profile release,
    # we must add a new entry to it with the new checksums, not overriding
    # the existing dict we have here. The rationale for that is to avoid
    # us forgetting to bump the profile version when modifying it.
    version_to_md5sum = {
        "v1": {
            "xenial": "350af6388522c8c28d8e00152fac98cc",
            "bionic": "b79ba7ea46882d35e6d10b08c7531f6f",
            "focal": "9ce4202e39d98c1499e3bce3c144e14f",
            "groovy": "05b1582d39237fb2d1b55c8782982bfd",
            "hirsute": "1f3851328bec6253f51b1f1dc9bcbf55",
        },
        "v2": {
            "xenial": "c4f83c97c2f39a39f1e997aa33e4bb66",
            "bionic": "0e35f88aa29c66374fbd9fe3b4a36257",
            "focal": "9ce4202e39d98c1499e3bce3c144e14f",
            "groovy": "05b1582d39237fb2d1b55c8782982bfd",
            "hirsute": "1f3851328bec6253f51b1f1dc9bcbf55",
            "impish": "c2a4e4d6a9c16f73f2f79fe34d3f638f",
        },
        "v3": {
            "xenial": "1f4d35dc74a550eb6458222a531a24c4",
            "bionic": "f0e13a4b8d11bc7b3d82c0f06ef72211",
            "default": "a740b8296455ba0b51ad093c77b0261b",
        },
    }

    @pytest.mark.parametrize("series", base_vm_profiles.keys())
    def test_profiles_md5sum_was_not_changed(self, series):
        """Test if the profiles md5sum still match.

        This test will ensure that the current profile version
        matches the md5sums we have stored for the profiles.
        """
        profiles_md5sum = self.version_to_md5sum[LXC_PROFILE_VERSION]

        current_profile_md5sum = hashlib.md5(
            base_vm_profiles[series].encode("utf-8")
        ).hexdigest()
        if series not in profiles_md5sum:
            pytest.fail(
                "Series {} md5sum not yet present: {}".format(
                    series, current_profile_md5sum
                )
            )
        profile_md5sum = profiles_md5sum[series]

        assert profile_md5sum == current_profile_md5sum



"""Test the config.py module."""
import os
from io import StringIO

import mock
import pytest

from pycloudlib.config import parse_config


def test_get_item_override():
    """Test the __getitem__ method override."""
    config = parse_config(StringIO(""))
    try:
        config["not_there"]
    except KeyError as e:
        assert str(e) == (
            "'not_there must be defined in pycloudlib.toml to make this call'"
        )


class TestParseConfig:
    """Test the parse_config function in config.py."""

    @mock.patch("toml.load")
    def test_argument_priority(self, m_load):
        """Test that config argument gets evaluated over files."""
        parse_config(StringIO(""))
        assert len(m_load.call_args_list) == 1
        assert "StringIO" in str(m_load.call_args_list)

    @mock.patch("toml.load")
    def test_env_var_priority(self, m_load):
        """Test that env var argument gets evaluated over files."""
        os.environ["PYCLOUDLIB_CONFIG"] = "/some/path"
        parse_config()
        assert len(m_load.call_args_list) == 1
        assert "/some/path" in str(m_load.call_args_list)

    @mock.patch("toml.load", side_effect=FileNotFoundError)
    def test_try_order(self, m_load):
        """Test order of config file checking."""
        os.environ["PYCLOUDLIB_CONFIG"] = "/some/path"
        with pytest.raises(ValueError):
            parse_config(StringIO(""))
        expected_order = [
            "StringIO",
            "/some/path",
            ".config/pycloudlib.toml",
            "/etc/pycloudlib.toml",
        ]
        for expected, actual in zip(expected_order, m_load.call_args_list):
            assert expected in str(actual)



"""LXD/LXD images' related functionalities."""
import functools
import itertools
import json
from typing import Any, List, Optional, Sequence, Tuple

from pycloudlib.util import subp

_REMOTE_DAILY = "ubuntu-daily"
_REMOTE_RELEASE = "ubuntu"


def find_last_fingerprint(
    daily: bool,
    release: str,
    is_container: bool,
    arch: str,
) -> Optional[str]:
    """Find last LXD image fingerprint.

    Args:
        daily: boolean, search on daily remote
        release: string, Ubuntu release to look for
        is_container: boolean, indicating whether it is container or not
        arch: string, architecture to use

    Returns:
        string, LXD fingerprint of latest image if found
    """
    remote = _REMOTE_DAILY if daily else _REMOTE_RELEASE
    remote += ":"
    base_filters = (
        ("architecture", arch),
        ("release", release),
        ("label", "daily" if daily else "release"),
    )
    filters = (
        *base_filters,
        ("type", "container" if is_container else "virtual-machine"),
    )
    find_images = functools.partial(_find_images, remote)
    found_images: Any = find_images(filters)

    # Note: This block is needed to support lxd <= 4
    if not found_images:
        if is_container:
            filters = (*base_filters, ("type", "squashfs"))
            found_images = iter(find_images(filters))
        else:
            filters_kvm = (*base_filters, ("type", "disk-kvm.img"))
            filters_disk1 = (*base_filters, ("type", "disk1.img"))
            filters_uefi1 = (*base_filters, ("type", "uefi1.img"))
            found_images = itertools.chain.from_iterable(
                map(
                    find_images,
                    (filters_kvm, filters_disk1, filters_uefi1),
                )
            )
    try:
        image = next(iter(found_images))
    except StopIteration:
        return None
    return "%s%s" % (remote, image["fingerprint"])


def find_release(image_id: str) -> Optional[str]:
    """Extract the base release from the image_id.

    Args:
        image_id: string, [<remote>:]<image identifier>, the image to
                  determine the release of

    Returns:
        A string containing the base release from the image_id or None if
        not found.
    """
    release = None
    images = _find_images(image_id)
    if not images:
        return release
    image_info = images[0]

    properties = image_info.get("properties")
    if not properties:
        return release

    os = properties.get("os")
    if not os:
        return release
    # images: images have "Ubuntu", ubuntu: images have "ubuntu"
    if os.lower() == "ubuntu":
        release = properties.get("release")

    return release


def find_image_serial(image_id: str) -> Optional[str]:
    """Find the image serial of a given LXD image.

    Args:
        image_id: LXD image fingerprint

    Returns:
        serial of latest image
    """
    if ":" not in image_id:
        remote = _REMOTE_DAILY
    else:
        remote, image_id = image_id.split(":", 1)
    filters = (("fingerprint", image_id),)
    image = _find_images(remote, filters)
    if not image:
        return None
    return image[0]["properties"]["serial"]


def _normalize_remote(remote: Optional[str] = None) -> str:
    if not remote:
        remote = _REMOTE_DAILY
    if ":" not in remote:
        remote += ":"
    return remote


def _find_images(
    remote: str, filters: Optional[Sequence[Tuple[str, str]]] = None
) -> List[dict]:
    """Find the images filtered by filters criteria.

    Args:
        remote: LXD's remote
        filters: Key-value filters to match on

    Returns:
        list of dictionaries with image's info
    """
    remote = _normalize_remote(remote)
    cmd = [
        "lxc",
        "image",
        "list",
        remote,
        "--format=json",
    ]
    if filters is not None:
        compiled_filters = map(lambda f: f"{f[0]}={f[1]}", filters)
        cmd.extend(compiled_filters)
    return json.loads(subp(cmd))



"""Openstack cloud type."""
import base64
from typing import Optional

import openstack

from pycloudlib.cloud import BaseCloud
from pycloudlib.config import ConfigFile
from pycloudlib.errors import (
    CloudSetupError,
    NetworkNotFoundError,
    PycloudlibError,
)
from pycloudlib.key import KeyPair
from pycloudlib.openstack.errors import OpenStackFlavorNotFound
from pycloudlib.openstack.instance import OpenstackInstance


class Openstack(BaseCloud):
    """Openstack cloud class."""

    _type = "openstack"

    def __init__(
        self,
        tag: str,
        timestamp_suffix: bool = True,
        config_file: Optional[ConfigFile] = None,
        *,
        network: Optional[str] = None,
    ):
        """Initialize the connection to openstack.

        Requires valid pre-configured environment variables or clouds.yaml.
        See https://docs.openstack.org/python-openstackclient/pike/configuration/index.html

        Args:
            tag: Name of instance
            timestamp_suffix: bool set True to append a timestamp suffix to the
                tag
            config_file: path to pycloudlib configuration file
            network: Name of the network to use (from openstack network list)

        """  # noqa: E501
        super().__init__(
            tag, timestamp_suffix, config_file, required_values=[network]
        )

        self.network = network or self.config["network"]
        self._openstack_keypair: Optional[KeyPair] = None
        self.conn = openstack.connect()

    def delete_image(self, image_id, **kwargs):
        """Delete an image.

        Args:
            image_id: string, id of the image to delete
        """
        self.conn.delete_image(image_id, wait=True)

    def released_image(self, release, **kwargs):
        """Not supported for openstack."""
        raise PycloudlibError(
            "Obtaining released image for a release is not supported on "
            "Openstack because we have no guarantee of what images will be "
            "available for any particular openstack setup."
        )

    def daily_image(self, release: str, **kwargs):
        """Not supported for openstack."""
        raise PycloudlibError(
            "Obtaining daily image for a release is not supported on "
            "Openstack because we have no guarantee of what images will be "
            "available for any particular openstack setup."
        )

    def image_serial(self, image_id):
        """Find the image serial of the latest daily image for a particular release.

        Args:
            image_id: string, Ubuntu image id

        Returns:
            string, serial of latest image

        """
        raise NotImplementedError

    def get_instance(self, instance_id, **kwargs) -> OpenstackInstance:
        """Get an instance by id.

        Args:
            instance_id: ID of instance to get

        Returns:
            An instance object to use to manipulate the instance further.

        """
        return OpenstackInstance(
            key_pair=self.key_pair,
            instance_id=instance_id,
            network_id=self._get_network_id(),
        )

    def _get_network_id(self):
        try:
            return self.conn.network.find_network(self.network).id
        except AttributeError as e:
            raise NetworkNotFoundError(resource_name=self.network) from e

    def launch(
        self,
        image_id,
        instance_type="m1.small",
        user_data="",
        wait=True,
        **kwargs,
    ) -> OpenstackInstance:
        """Launch an instance.

        Args:
            image_id: string, image ID to use for the instance
            instance_type: string, type (flavor) of instance to create
            user_data: used by cloud-init to run custom scripts/configuration
            wait: wait for instance to be live
            **kwargs: dictionary of other arguments to pass to launch

        Returns:
            An instance object to use to manipulate the instance further.
        Raises: ValueError on invalid image_id
        """
        if not image_id:
            raise ValueError(
                f"{self._type} launch requires image_id param."
                f" Found: {image_id}"
            )
        network_id = self._get_network_id()
        networks = [{"uuid": network_id}]
        if not self._openstack_keypair:
            self._openstack_keypair = self._get_openstack_keypair()
        if user_data:
            user_data = base64.b64encode(user_data.encode()).decode()
        else:
            user_data = ""

        flavor = self.conn.compute.find_flavor(instance_type)
        if flavor is None:
            raise OpenStackFlavorNotFound(
                "No Openstack flavor found named {}. Please pass a valid "
                "Openstack flavor as the `instance_type` when calling "
                "launch.".format(instance_type)
            )

        instance = self.conn.compute.create_server(
            name=self.tag,
            image_id=image_id,
            flavor_id=flavor.id,
            networks=networks,
            key_name=self._openstack_keypair.name,
            user_data=user_data,
            wait=wait,
            **kwargs,
        )
        instance = OpenstackInstance(
            key_pair=self.key_pair,
            instance_id=instance.id,
            network_id=network_id,
            connection=self.conn,
        )
        if wait:
            instance.wait()
        return instance

    def snapshot(self, instance, clean=True, **kwargs):
        """Snapshot an instance and generate an image from it.

        Args:
            instance: Instance to snapshot
            clean: run instance clean method before taking snapshot

        Returns:
            An image id

        """
        if clean:
            instance.clean()
        instance.shutdown()
        image = self.conn.create_image_snapshot(
            "{}-snapshot".format(self.tag), instance.server.id, wait=True
        )
        return image.id

    def use_key(self, public_key_path, private_key_path=None, name=None):
        """Use an existing key.

        Args:
            public_key_path: path to the public key to upload
            private_key_path: path to the private key
            name: name to reference key by
        """
        super().use_key(public_key_path, private_key_path, name)
        self._openstack_keypair = self._get_openstack_keypair()

    def _get_openstack_keypair(self) -> KeyPair:
        """Get openstack keypair corresponding to this instances keypair.

        When creating an openstack instance, a keypair (maintained in
        openstack) must be created first. This method gets or creates
        the openstack keypair corresponding to the keypair already created
        for this cloud instance.
        """
        name = self.key_pair.name
        public_key_content = self.key_pair.public_key_content

        openstack_keypair = self.conn.get_keypair(name)
        if not openstack_keypair:
            # If the openstack keypair doesn't exist, create it
            return self.conn.create_keypair(name, public_key_content)
        if public_key_content == openstack_keypair.public_key:
            return openstack_keypair
        raise CloudSetupError(
            "An openstack keypair with name {name} already exists, but its"
            " public key doesn't match the public key passed in.\n"
            "{name} key: {openstack_key}\n"
            "Passed in key: {passed_key}".format(
                name=name,
                openstack_key=openstack_keypair.public_key,
                passed_key=public_key_content,
            )
        )



# This file is part of pycloudlib. See LICENSE file for license information.
"""Helpers for shell string and processing."""

import base64
import collections.abc
import datetime
import os
import platform
import re
import shlex
import subprocess
import tempfile
from errno import ENOENT
from typing import Dict
from urllib.parse import parse_qs, urlparse

from pycloudlib.result import Result

UBUNTU_RELEASE_VERSION_MAP = {
    "lunar": "23.04",
    "kinetic": "22.10",
    "jammy": "22.04",
    "focal": "20.04",
    "bionic": "18.04",
    "xenial": "16.04",
}

LTS_RELEASES = ["xenial", "bionic", "focal", "jammy"]


def chmod(path, mode):
    """Run chmod on a file or directory.

    Args:
        path: string of path to run on
        mode: int of mode to apply
    """
    real_mode = _safe_int(mode)
    if path and real_mode:
        os.chmod(path, real_mode)


def is_writable_dir(path):
    """Make sure dir is writable.

    Args:
        path: path to determine if writable

    Returns:
        boolean with result

    """
    try:
        touch(path)
        os.remove(tempfile.mkstemp(dir=os.path.abspath(path))[1])
    except (IOError, OSError):
        return False
    return True


def mkdtemp(prefix="pycloudlib"):
    """Make a temporary directory.

    Args:
        prefix: optional, temproary dir name prefix (default: pycloudlib)

    Returns:
        tempfile object that was created

    """
    return tempfile.mkdtemp(prefix=prefix)


def rmfile(path):
    """Delete a file.

    Args:
        path: run unlink on specific path
    """
    try:
        os.unlink(path)
    except OSError as error:
        if error.errno != ENOENT:
            raise error


def shell_pack(cmd):
    """Return a string that can shuffled through 'sh' and execute cmd.

    In Python subprocess terms:
        check_output(cmd) == check_output(shell_pack(cmd), shell=True)

    Args:
        cmd: list or string of command to pack up

    Returns:
        base64 encoded string

    """
    if isinstance(cmd, str):
        cmd = [cmd]
    else:
        cmd = list(cmd)

    stuffed = shell_safe(cmd)
    # for whatever reason b64encode returns bytes when it is clearly
    # representable as a string by nature of being base64 encoded.
    b64 = base64.b64encode(stuffed.encode()).decode()
    return 'eval set -- "$(echo %s | base64 --decode)" && exec "$@"' % b64


def shell_quote(cmd):
    """Quote a shell string.

    Args:
        cmd: command to quote

    Returns:
        quoted string

    """
    if isinstance(cmd, (tuple, list)):
        return " ".join([shlex.quote(x) for x in cmd])

    return shlex.quote(cmd)


def shell_safe(cmd):
    """Produce string safe shell string.

    Create a string that can be passed to $(set -- <string>) to produce
    the same array that cmd represents.

    Internally we utilize 'getopt's ability/knowledge on how to quote
    strings to be safe for shell.  This implementation could be changed
    to be pure python.  It is just a matter of correctly escaping
    or quoting characters like: ' " ^ & $ ; ( ) ...

    Args:
        cmd: command as a list

    Returns:
        shell safe string

    """
    out = subprocess.check_output(
        ["getopt", "--shell", "sh", "--options", "", "--", "--"] + list(cmd)
    )

    # out contains ' -- <data>\n'. drop the ' -- ' and the '\n'
    return out.decode()[4:-1]


def subp(
    args, data=None, env=None, shell=False, rcs=(0,), shortcircuit_stdin=True
):
    """Subprocess wrapper.

    Args:
        args: command to run
        data: data to pass
        env: optional env to use
        shell: optional shell to use
        rcs: tuple of successful exit codes, default: (0)
        shortcircuit_stdin: bind stdin to /dev/null if no data is given

    Returns:
        Tuple of out, err, return_code

    """
    devnull_fp = None

    if data is not None:
        stdin = subprocess.PIPE
        if not isinstance(data, bytes):
            data = data.encode()
    elif shortcircuit_stdin:
        # using devnull assures any reads get null, rather
        # than possibly waiting on input.
        devnull_fp = open(os.devnull, "rb")  # pylint: disable=R1732
        stdin = devnull_fp
    else:
        stdin = None

    bytes_args = _convert_args(args)

    try:
        process = subprocess.Popen(  # pylint: disable=R1732
            bytes_args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=stdin,
            env=env,
            shell=shell,
        )
        (out, err) = process.communicate(data)
    finally:
        if devnull_fp:
            devnull_fp.close()

    rc = process.returncode
    out = "" if not out else out.rstrip().decode("utf-8")
    err = "" if not err else err.rstrip().decode("utf-8")

    if rcs and rc not in rcs:
        if err:
            errmsg = err
        elif out:
            errmsg = out
        else:
            errmsg = "command failed silently"
        errmsg = "Failure (rc=%s): %s" % (rc, errmsg)
        raise RuntimeError(errmsg)

    return Result(out, err, rc)


def touch(path, mode=None):
    """Ensure a directory exists with a specific mode, it not create it.

    Args:
        path: path to directory to create
        mode: optional, mode to set directory to
    """
    if not os.path.isdir(path):
        with os.path.dirname(path):
            os.makedirs(path)
        chmod(path, mode)
    else:
        chmod(path, mode)


def _get_local_ubuntu_arch() -> str:
    """Return the Ubuntu architecture suitable for the local system.

    This is not simply the local machine hardware name, as in some cases it
    differs from the Ubuntu architecture name. The most common case is x86_64
    hardware, for which the Ubuntu architecture name is 'amd64'. This function
    implements the required mapping.

    On Debian and Ubuntu systmes the full mapping between the GNU architecture
    names and the Ubuntu (Debian) architecture names is available in the
    '/usr/share/dpkg/cputable' file, however the GNU architecture names are
    again different from the machine hardware names from e.g. uname(1) or
    os.uname(). The full mapping is available in the 'config.guess' script from
    the GNU autotools, and it's complex. Let's keep it simple here, mapping
    only what is relevant for Ubuntu.
    """
    arch_map = dict(
        i686="i386",
        x86_64="amd64",
        aarch64="arm64",
        ppc="powerpc",
        ppc64el="ppc64el",
        ppcle="powerpcel",
    )

    local_arch = platform.machine()
    local_ubuntu_arch = arch_map.get(local_arch, local_arch)

    return local_ubuntu_arch


def _convert_args(args):
    """Convert subp arguments to bytes.

    Popen converts entries in the arguments array from non-bytes to bytes.
    When locale is unset it may use ascii for that encoding which can
    cause UnicodeDecodeErrors. (LP: #1751051)

    Args:
        args: string, bytes, or list of arguments to convert to bytes

    Returns:
        byte argument list

    """
    if isinstance(args, bytes):
        bytes_args = args
    elif isinstance(args, str):
        bytes_args = args.encode("utf-8")
    else:
        bytes_args = [
            x if isinstance(x, bytes) else x.encode("utf-8") for x in args
        ]

    return bytes_args


def _safe_int(possible_int):
    """Create an int as safely as possbile.

    Args:
        possible_int: variable to create into a integer

    Returns:
        integration or None

    """
    try:
        return int(possible_int)
    except (ValueError, TypeError):
        return None


def get_timestamped_tag(tag):
    """Create tag with current timestamp.

    Args:
        tag: string, Base tag to be used

    Returns
        An updated tag with current timestamp

    """
    return "%s-%s" % (tag, datetime.datetime.now().strftime("%m%d-%H%M%S"))


def validate_tag(tag):
    """Ensure tag will work as name for clouds that use it."""
    # Currently google is the most restrictive, so just use that
    # regex verbatum. You can trigger the error message that contains
    # this regex by attempting to create an instance with a name
    # of '-'
    regex = r"^(?:[a-z](?:[-a-z0-9]{0,61}[a-z0-9])?)$"
    if not re.match(regex, tag):
        raise ValueError(
            "Invalid tag specified. After being timestamped, "
            "tag must pass regex.\n"
            "Regex: {}\n"
            "Tag  : {}".format(regex, tag)
        )
    return tag


def update_nested(mapping, update):
    """Update mapping with update value at given update key.

    Example:
      original_dict = {'a': {'b': {'c': 'd'}}}
      update = {'a': {'b': {'c': 'e'}}}
      update_nested(original_dict, update)
      original_dict == {'a': {'b': {'c': 'e'}}}
    """
    for key, value in update.items():
        if isinstance(value, collections.abc.Mapping):
            mapping[key] = update_nested(mapping.get(key, {}), value)
        else:
            mapping[key] = value
    return mapping


def get_query_params(uri: str) -> Dict[str, list]:
    """Extract query params from `uri`.

    >>> url = "https://cloud.com/v1/vpcs?limit=1&start=r134-fe06d70f"
    >>> get_query_params(url)
    {'limit': ['1'], 'start': ['r134-fe06d70f']}
    """
    return parse_qs(urlparse(uri).query)


def get_query_param(uri: str, param: str) -> list:
    """Extract query query params of `param` from `uri`.

    >>> url = "https://cloud.com/v1/vpcs?limit=1&start=r134-fe06d70f"
    >>> get_query_param(url, "start")
    ['r134-fe06d70f']
    """
    return get_query_params(uri).get(param, [])



import ipaddress
from contextlib import contextmanager, suppress
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Generator

import pytest

import pycloudlib
from pycloudlib.cloud import BaseCloud
from pycloudlib.instance import BaseInstance
from pycloudlib.util import LTS_RELEASES

cloud_config = """\
#cloud-config
runcmd:
  - echo 'hello' >> /var/tmp/example.txt
"""


@pytest.fixture
def cloud(request):
    cloud_instance: BaseCloud = request.param(
        tag="pycl-test",
        timestamp_suffix=True,
    )

    if isinstance(cloud_instance, pycloudlib.EC2):
        cloud_instance.upload_key(
            public_key_path=cloud_instance.config["public_key_path"],
            private_key_path=cloud_instance.config["private_key_path"],
            name=cloud_instance.tag,
        )

    yield cloud_instance

    if isinstance(cloud_instance, pycloudlib.EC2):
        cloud_instance.delete_key(name=cloud_instance.tag)


@contextmanager
def launch_instance(
    cloud_instance: BaseCloud, **kwargs
) -> Generator[BaseInstance, None, None]:
    instance = cloud_instance.launch(**kwargs)
    try:
        yield instance
    finally:
        instance.delete()


def assert_example_output(instance: BaseInstance):
    example_output = instance.execute("cat /var/tmp/example.txt").stdout
    assert example_output == "hello"


def exercise_push_pull(instance: BaseInstance):
    with TemporaryDirectory() as tmpdir:
        push_path = Path(tmpdir).joinpath("pushed")
        push_path.write_text("pushed", encoding="utf-8")
        instance.push_file(str(push_path), "/var/tmp/pushed")
        assert (
            "pushed" == instance.execute("cat /var/tmp/pushed").stdout.strip()
        )

        instance.execute("echo 'pulled' > /var/tmp/pulled")
        pull_path = Path(tmpdir).joinpath("pulled")
        instance.pull_file("/var/tmp/pulled", str(pull_path))
        assert pull_path.read_text(encoding="utf-8").strip() == "pulled"


def exercise_instance(instance: BaseInstance):
    assert instance.name is not None
    assert ipaddress.ip_address(instance.ip)

    with suppress(NotImplementedError):
        ip_address = instance.add_network_interface()
        try:
            assert ipaddress.ip_address(ip_address)
        finally:
            instance.remove_network_interface(ip_address)

    exercise_push_pull(instance)

    assert_example_output(instance)
    boot_id = instance.get_boot_id()
    instance.restart()
    assert boot_id != instance.get_boot_id()

    assert_example_output(instance)
    instance.shutdown()

    instance.start()
    assert_example_output(instance)


@pytest.mark.parametrize(
    "cloud",
    [
        pytest.param(
            pycloudlib.Azure, id="azure", marks=pytest.mark.main_check
        ),
        pytest.param(pycloudlib.EC2, id="ec2", marks=pytest.mark.main_check),
        pytest.param(pycloudlib.GCE, id="gce", marks=pytest.mark.main_check),
        pytest.param(pycloudlib.IBM, id="ibm"),
        pytest.param(
            pycloudlib.LXDContainer, id="lxd_container", marks=pytest.mark.ci
        ),
        pytest.param(pycloudlib.LXDVirtualMachine, id="lxd_vm"),
        pytest.param(pycloudlib.OCI, id="oci"),
        # For openstack we first need a reliable way of obtaining the
        # image id
        # pytest.param(pycloudlib.Openstack, id="openstack"),
    ],
    indirect=True,
)
def test_public_api(cloud: BaseCloud):
    """Shallow test of (most) public functions in the base API."""
    latest_lts = LTS_RELEASES[-1]
    print(f"Using Ubuntu {latest_lts} release")
    try:
        image_id = cloud.released_image(release=latest_lts)
    except NotImplementedError:
        image_id = cloud.daily_image(release=latest_lts)
    with suppress(NotImplementedError):
        # Not sure there's a great way to test this other than not raising
        cloud.image_serial(image_id)

    with launch_instance(
        cloud, image_id=image_id, user_data=cloud_config, wait=False
    ) as instance:
        instance.wait()
        exercise_instance(instance)

        instance.clean()
        result = instance.execute("sudo rm /var/tmp/example.txt")
        snapshot_id = cloud.snapshot(instance)

    try:
        with launch_instance(
            cloud, image_id=snapshot_id, user_data=cloud_config, wait=False
        ) as instance_from_snapshot:
            instance_from_snapshot.wait()
            exercise_instance(instance_from_snapshot)
    finally:
        cloud.delete_image(snapshot_id)



# This file is part of pycloudlib. See LICENSE file for license information.
"""GCE Cloud type.

This is an initial implimentation of the GCE class. It enables
authentication into the cloud, finding an image, and launching an
instance. It however, does not allow any further actions from occuring.
"""

import logging
import os
import time
from itertools import count
from typing import Optional

import googleapiclient.discovery

from pycloudlib.cloud import BaseCloud, ImageType
from pycloudlib.config import ConfigFile
from pycloudlib.errors import (
    CloudSetupError,
    ImageNotFoundError,
    PycloudlibError,
)
from pycloudlib.gce.instance import GceInstance
from pycloudlib.gce.util import get_credentials, raise_on_error
from pycloudlib.util import UBUNTU_RELEASE_VERSION_MAP, subp

logging.getLogger("googleapiclient.discovery").setLevel(logging.WARNING)


class GCE(BaseCloud):
    """GCE Cloud Class."""

    _type = "gce"

    def __init__(
        self,
        tag: str,
        timestamp_suffix: bool = True,
        config_file: Optional[ConfigFile] = None,
        *,
        credentials_path: Optional[str] = None,
        project: Optional[str] = None,
        region: Optional[str] = None,
        zone: Optional[str] = None,
        service_account_email: Optional[str] = None,
    ):
        """Initialize the connection to GCE.

        Args:
            tag: string used to name and tag resources with
            timestamp_suffix: bool set True to append a timestamp suffix to the
                tag
            config_file: path to pycloudlib configuration file
            credentials_path: path to credentials file for GCE
            project: GCE project
            region: GCE region
            zone: GCE zone
            service_account_email: service account to bind launched
                                   instances to
        """
        super().__init__(
            tag,
            timestamp_suffix,
            config_file,
            required_values=[credentials_path, project],
        )

        self._log.debug("logging into GCE")

        self.credentials_path = ""
        if credentials_path:
            self.credentials_path = credentials_path
        elif "GOOGLE_APPLICATION_CREDENTIALS" in os.environ:
            self.credentials_path = os.environ[
                "GOOGLE_APPLICATION_CREDENTIALS"
            ]
        elif "credentials_path" in self.config:
            self.credentials_path = os.path.expandvars(
                os.path.expanduser(self.config["credentials_path"])
            )

        credentials = get_credentials(self.credentials_path)

        if not project:
            if "project" in self.config:
                project = self.config["project"]
            elif "GOOGLE_CLOUD_PROJECT" in os.environ:
                project = os.environ["GOOGLE_CLOUD_PROJECT"]
            else:
                command = ["gcloud", "config", "get-value", "project"]
                exception_text = (
                    "Could not obtain GCE project id. Set it in the "
                    "pycloudlib config or setup the gcloud cli."
                )
                try:
                    result = subp(command, rcs=())
                except FileNotFoundError as e:
                    raise CloudSetupError(exception_text) from e
                if not result.ok:
                    exception_text += "\nstdout: {}\nstderr: {}".format(
                        result.stdout, result.stderr
                    )
                    raise CloudSetupError(exception_text)
                project = result.stdout

        # disable cache_discovery due to:
        # https://github.com/google/google-api-python-client/issues/299
        self.compute = googleapiclient.discovery.build(
            "compute",
            "v1",
            cache_discovery=False,
            credentials=credentials,
        )
        region = region or self.config.get("region") or "us-west2"
        zone = zone or self.config.get("zone") or "a"
        self.project = project
        self.region = region
        self.zone = "%s-%s" % (region, zone)
        self.instance_counter = count()
        self.service_account_email = service_account_email or self.config.get(
            "service_account_email"
        )

    def released_image(
        self, release, *, image_type: ImageType = ImageType.GENERIC, **kwargs
    ):
        """ID of the latest released image for a particular release.

        Args:
            release: The release to look for

        Returns:
            A single string with the latest released image ID for the
            specified release.
        """
        return self.daily_image(release=release, image_type=image_type)

    def _get_project(self, image_type: ImageType):
        return (
            "ubuntu-os-cloud-devel"
            if image_type == ImageType.GENERIC
            else "ubuntu-os-pro-cloud"
        )

    def _get_name_filter(self, release: str, image_type: ImageType):
        if image_type == ImageType.GENERIC:
            return "daily-ubuntu-{}-{}-*".format(
                UBUNTU_RELEASE_VERSION_MAP[release].replace(".", ""), release
            )

        if image_type == ImageType.PRO:
            return "ubuntu-pro-{}-{}-*".format(
                UBUNTU_RELEASE_VERSION_MAP[release].replace(".", ""), release
            )

        if image_type == ImageType.PRO_FIPS:
            return "ubuntu-pro-fips-{}-{}-*".format(
                UBUNTU_RELEASE_VERSION_MAP[release].replace(".", ""), release
            )

        raise ValueError("Invalid image_type: {}".format(image_type.value))

    def _query_image_list(
        self, release: str, project: str, name_filter: str, arch: str
    ):
        """Query full list of images.

        image list API docs:
        https://googleapis.github.io/google-api-python-client/docs/dyn/compute_v1.images.html#list

        The image list API doesn't allow filtering and sorting in one request
        so we need to do one of those locally.
        Filtering via the API results in fewer requests on average than
        sorting via the API.
        So we filter via the API and loop through all pages to get the full
        image list matching that filter.
        500 is the maximum allowed page size
        Then we can sort locally and grab the latest image.

        Args:
            release: string, Ubuntu release to look for
            project: GCE project
            name_filter: name to filter with
            arch: images' architecture

        Returns:
            list of images matching the given filters
        """
        filter_string = "(name={}) AND (architecture={})".format(
            name_filter, arch.upper()
        )

        # SPECIAL CASE
        # Google didn't start including architecture in image descriptions
        # until after xenial stopped getting published
        # All xenial images are x86_64, so:
        #   1. we can return early for non-x86_64 xenial queries
        #   2. for xenial + x86_64 we don't include the architecture in the
        #      filter
        if release == "xenial":
            if arch != "x86_64":
                return []
            filter_string = "name={}".format(name_filter)

        image_list = []
        page_token = ""
        reqs = 0
        while page_token is not None:
            image_list_result = (
                self.compute.images()
                .list(
                    project=project,
                    filter=filter_string,
                    maxResults=500,
                    pageToken=page_token,
                )
                .execute()
            )
            reqs += 1
            image_list += image_list_result.get("items", [])
            page_token = image_list_result.get("nextPageToken", None)

        self._log.debug(
            (
                'Fetched entire image list (%i results) matching "%s" in %i'
                " requests"
            ),
            len(image_list),
            filter_string,
            reqs,
        )

        return image_list

    def daily_image(
        self,
        release: str,
        *,
        arch: str = "x86_64",
        image_type: ImageType = ImageType.GENERIC,
        **kwargs,
    ):
        """Find the id of the latest image for a particular release.

        Args:
            release: string, Ubuntu release to look for

        Returns:
            string, path to latest daily image

        """
        self._log.debug(
            "finding daily Ubuntu image for arch: %s and release: %s",
            arch,
            release,
        )
        project = self._get_project(image_type=image_type)
        name_filter = self._get_name_filter(
            release=release, image_type=image_type
        )

        image_list = self._query_image_list(
            release, project, name_filter, arch
        )

        if not image_list:
            msg = (
                "Could not find {} image for arch: {} and release: {}".format(
                    image_type.value,
                    arch,
                    release,
                )
            )
            self._log.warning(msg)
            raise ImageNotFoundError(
                image_type=image_type.value, arch=arch, release=release
            )

        image = sorted(image_list, key=lambda x: x["creationTimestamp"])[-1]
        self._log.debug(
            'Found image name "%s" for arch "%s"',
            image["name"],
            arch,
        )
        return "projects/{}/global/images/{}".format(project, image["id"])

    def image_serial(self, image_id):
        """Find the image serial of the latest daily image for a particular release.

        Args:
            image_id: string, Ubuntu image id

        Returns:
            string, serial of latest image

        """
        raise NotImplementedError

    def delete_image(self, image_id, **kwargs):
        """Delete an image.

        Args:
            image_id: string, id of the image to delete
        """
        api_image_id = (
            self.compute.images()
            .get(project=self.project, image=os.path.basename(image_id))
            .execute()["id"]
        )
        response = (
            self.compute.images()
            .delete(
                project=self.project,
                image=api_image_id,
            )
            .execute()
        )

        raise_on_error(response)

    def get_instance(self, instance_id, name=None):
        """Get an instance by id.

        Args:
            instance_id: The instance ID returned upon creation

        Returns:
            An instance object to use to manipulate the instance further.

        """
        return GceInstance(
            self.key_pair,
            instance_id,
            self.project,
            self.zone,
            self.credentials_path,
            name=name,
        )

    def launch(
        self,
        image_id,
        instance_type="n1-standard-1",
        user_data=None,
        wait=True,
        **kwargs,
    ):
        """Launch instance on GCE and print the IP address.

        Args:
            image_id: string, image ID for instance to use
            instance_type: string, instance type to launch
            user_data: string, user-data to pass to instance
            wait: boolean, wait for instance to come up
            kwargs: other named arguments to add to instance JSON
        Raises: ValueError on invalid image_id
        """
        if not image_id:
            raise ValueError(
                f"{self._type} launch requires image_id param."
                f" Found: {image_id}"
            )
        instance_name = "i{}-{}".format(next(self.instance_counter), self.tag)
        config = {
            "name": instance_name,
            "machineType": "zones/%s/machineTypes/%s"
            % (self.zone, instance_type),
            "disks": [
                {
                    "boot": True,
                    "autoDelete": True,
                    "initializeParams": {
                        "sourceImage": image_id,
                    },
                }
            ],
            "networkInterfaces": [
                {
                    "network": "global/networks/default",
                    "accessConfigs": [
                        {"type": "ONE_TO_ONE_NAT", "name": "External NAT"}
                    ],
                }
            ],
            "metadata": {
                "items": [
                    {
                        "key": "ssh-keys",
                        "value": "ubuntu:%s"
                        % self.key_pair.public_key_content,
                    }
                ]
            },
        }

        if self.service_account_email:
            config["serviceAccounts"] = [{"email": self.service_account_email}]

        if user_data:
            user_metadata = {"key": "user-data", "value": user_data}
            config["metadata"]["items"].append(user_metadata)

        operation = (
            self.compute.instances()
            .insert(project=self.project, zone=self.zone, body=config)
            .execute()
        )
        raise_on_error(operation)

        result = (
            self.compute.instances()
            .get(
                project=self.project,
                zone=self.zone,
                instance=instance_name,
            )
            .execute()
        )
        raise_on_error(result)

        instance = self.get_instance(result["id"], name=result["name"])
        if wait:
            self._wait_for_operation(operation, operation_type="zone")
            instance.wait()

        return instance

    def snapshot(self, instance: GceInstance, clean=True, **kwargs):
        """Snapshot an instance and generate an image from it.

        Args:
            instance: Instance to snapshot
            clean: run instance clean method before taking snapshot

        Returns:
            An image id
        """
        response = (
            self.compute.disks()
            .list(project=self.project, zone=self.zone)
            .execute()
        )

        instance_disks = [
            disk for disk in response["items"] if disk["name"] == instance.name
        ]

        if len(instance_disks) > 1:
            raise PycloudlibError(
                "Snapshotting an image with multiple disks not supported"
            )

        instance.shutdown()

        snapshot_name = "{}-image".format(instance.name)
        operation = (
            self.compute.images()
            .insert(
                project=self.project,
                body={
                    "name": snapshot_name,
                    "sourceDisk": instance_disks[0]["selfLink"],
                },
            )
            .execute()
        )
        raise_on_error(operation)
        self._wait_for_operation(operation)

        return "projects/{}/global/images/{}".format(
            self.project, snapshot_name
        )

    def _wait_for_operation(
        self, operation, operation_type="global", sleep_seconds=300
    ):
        response = None
        kwargs = {"project": self.project, "operation": operation["name"]}
        if operation_type == "zone":
            kwargs["zone"] = self.zone
            api = self.compute.zoneOperations()
        else:
            api = self.compute.globalOperations()
        for _ in range(sleep_seconds):
            try:
                response = api.get(**kwargs).execute()
            except ConnectionResetError:
                # This exception is known to be raised by GCE every so often:
                # https://github.com/canonical/pycloudlib/issues/101.
                response = {
                    "status": "ConnectionResetError",
                    "statusMessage": "n/a",
                }
            else:
                if response["status"] == "DONE":
                    break
            time.sleep(1)
        else:
            raise PycloudlibError(
                "Expected DONE state, but found {} after waiting {} seconds. "
                "Check GCE console for more details. \n"
                "Status message: {}".format(
                    response["status"],
                    sleep_seconds,
                    response["statusMessage"],
                )
            )



#!/usr/bin/env python3
"""Python packaging configuration."""
import os
from pathlib import Path

from setuptools import find_packages, setup


def read_readme():
    """Read and return text of README.md."""
    pwd = os.path.abspath(os.path.dirname(__name__))
    readme_file = os.path.join(pwd, "README.md")
    with open(readme_file, "r", encoding="utf-8") as readme:
        readme_txt = readme.read()

    return readme_txt


def read_version():
    """Read and return text of VERSION."""
    return (
        Path(__file__)
        .parent.joinpath("VERSION")
        .read_text(encoding="utf-8")
        .strip()
    )


INSTALL_REQUIRES = [
    "azure-cli-core >= 2.21.0",
    "azure-identity",
    "azure-mgmt-compute >= 17",
    "azure-mgmt-network >= 16",
    "azure-mgmt-resource >= 15",
    "boto3 >= 1.14.20",
    "botocore >= 1.17.20",
    "google-api-python-client >= 1.7.7",
    "ibm-platform-services",
    "knack >= 0.7.1",
    "oci >= 2.17.0",
    "paramiko >= 2.9.2",
    "protobuf < 3.20.0",
    "pyparsing >= 2, < 3.0.0",
    "python-openstackclient >= 5.2.1",
    "pyyaml >= 5.1",
    "requests >= 2.22",
    "toml == 0.10",
]

EXTRAS_REQUIRE = {
    ":python_version == '3.6'": [
        "ibm-cloud-sdk-core == 3.14.0",  # last py36 compatible version
        "ibm-vpc == 0.10",
    ],
    ":python_version >= '3.7'": [
        "ibm-cloud-sdk-core >= 3.14.0",
        "ibm-vpc >= 0.10",
    ],
}

setup(
    name="pycloudlib",
    version=read_version(),
    description=(
        "Python library to launch, interact, and snapshot cloud instances"
    ),
    long_description=read_readme(),
    long_description_content_type="text/markdown",
    author="pycloudlib-devs",
    author_email="pycloudlib-devs@lists.launchpad.net",
    url="https://launchpad.net/pycloudlib",
    license="GNU General Public License v3 (GPLv3)",
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    zip_safe=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
)



# This file is part of pycloudlib. See LICENSE file for license information.
"""Azure instance."""

import time

from pycloudlib.errors import PycloudlibTimeoutError
from pycloudlib.instance import BaseInstance


class AzureInstance(BaseInstance):
    """Azure backed instance."""

    _type = "azure"

    def __init__(self, key_pair, client, instance):
        """Set up instance.

        Args:
            key_pair: SSH key object
            client: Azure compute management client
            instance: created azure instance object
        """
        super().__init__(key_pair)

        self._client = client
        self._instance = instance
        self.boot_timeout = 300
        self.status = "active"

    def wait_for_delete(self):
        """Wait for instance to be deleted."""
        raise NotImplementedError

    def wait_for_stop(self):
        """Wait for instance stop."""
        for _ in range(100):
            power_state = (
                self._client.virtual_machines.get(
                    resource_group_name=self._instance["rg_name"],
                    vm_name=self.name,
                    expand="instanceView",
                )
                .instance_view.statuses[1]
                .display_status
            )
            if power_state == "VM stopped":
                return
            time.sleep(1)
        raise PycloudlibTimeoutError

    @property
    def image_id(self):
        """Return the image_id from which this instance was created."""
        storage_profile = (
            self._instance["vm"].as_dict().get("storage_profile", {})
        )
        image_ref = storage_profile.get("image_reference", {})

        if image_ref:
            return ":".join(
                [
                    image_ref.get("publisher", "").lower(),
                    image_ref.get("offer", ""),
                    image_ref.get("sku", ""),
                    image_ref.get("version", ""),
                ]
            )

        # Snapshot instances will not contain such info. For them, we will
        # return a default string
        return "snapshot-image"

    @property
    def ip(self):
        """Return IP address of instance."""
        return self._instance["ip_address"]

    @property
    def id(self):
        """Return instance id."""
        return self._instance["vm"].id

    @property
    def name(self):
        """Return instance name."""
        return self._instance["vm"].name

    @property
    def sku(self):
        """Return instance sku."""
        image_profile = self._instance["vm"].storage_profile.image_reference
        return getattr(image_profile, "sku", "")

    @property
    def offer(self):
        """Return instance sku."""
        image_profile = self._instance["vm"].storage_profile.image_reference
        return getattr(image_profile, "offer", "")

    def shutdown(self, wait=True, **kwargs):
        """Shutdown the instance.

        Args:
            wait: wait for the instance shutdown
        """
        shutdown = self._client.virtual_machines.begin_power_off(
            resource_group_name=self._instance["rg_name"], vm_name=self.name
        )

        if wait:
            shutdown.wait()

    def generalize(self):
        """Set the OS state of the instance to generalized."""
        self._client.virtual_machines.generalize(
            resource_group_name=self._instance["rg_name"], vm_name=self.name
        )

    def start(self, wait=True):
        """Start the instance.

        Args:
            wait: wait for the instance to start.
        """
        start = self._client.virtual_machines.begin_start(
            resource_group_name=self._instance["rg_name"], vm_name=self.name
        )

        if wait:
            start.wait()
            self.wait()

    def _do_restart(self, **kwargs):
        """Restart the instance."""
        self._client.virtual_machines.begin_restart(
            resource_group_name=self._instance["rg_name"], vm_name=self.name
        )

    def delete(self, wait=True):
        """Delete instance."""
        delete = self._client.virtual_machines.begin_delete(
            resource_group_name=self._instance["rg_name"], vm_name=self.name
        )

        if wait:
            delete.wait()

        self.status = "deleted"



"""Tests related to pycloudlib.instance module."""
from itertools import repeat
from unittest import mock

import pytest
from paramiko import SSHException

from pycloudlib.errors import PycloudlibTimeoutError
from pycloudlib.instance import BaseInstance
from pycloudlib.result import Result

# Disable this pylint check as fixture usage incorrectly triggers it:
# pylint: disable=redefined-outer-name


@pytest.fixture
def concrete_instance_cls():
    """Return a BaseInstance subclass which can be instantiated.

    Source: https://stackoverflow.com/a/28738073
    """
    with mock.patch.object(BaseInstance, "__abstractmethods__", set()):
        yield BaseInstance


class TestWait:
    """Tests covering pycloudlib.instance.Instance.wait."""

    def test_wait(self, concrete_instance_cls):
        """Test wait calls the two methods it should with correct passthrough.

        (`None` is used to test the default.)
        """
        instance = concrete_instance_cls(key_pair=None)
        with mock.patch.multiple(
            instance,
            _wait_for_instance_start=mock.DEFAULT,
            _wait_for_execute=mock.DEFAULT,
            _wait_for_cloudinit=mock.DEFAULT,
        ) as mocks:
            instance.wait()

        assert 1 == mocks["_wait_for_instance_start"].call_count
        assert 1 == mocks["_wait_for_execute"].call_count
        assert 1 == mocks["_wait_for_cloudinit"].call_count

    @pytest.mark.parametrize(
        "execute_effect",
        [lambda *args, **kwargs: Result("", "", 1), SSHException],
    )
    @mock.patch.object(BaseInstance, "execute")
    @mock.patch("pycloudlib.instance.time.sleep")
    @mock.patch("pycloudlib.instance.time.time")
    def test_wait_execute_failure(
        self, m_time, m_sleep, m_execute, execute_effect, concrete_instance_cls
    ):
        """Test wait calls when execute command fails."""
        instance = concrete_instance_cls(key_pair=None)
        m_time.side_effect = [1, 2, 40 * 60, 40 * 60 + 1]
        m_execute.side_effect = execute_effect
        expected_msg = (
            "Instance can't be reached after 40 minutes. "
            "Failed to obtain new boot id"
        )
        expected_call_args = [
            mock.call("cat /proc/sys/kernel/random/boot_id", no_log=True)
        ] * 2

        with pytest.raises(PycloudlibTimeoutError) as excinfo:
            instance.wait()

        assert expected_msg == str(excinfo.value)
        assert m_sleep.call_count == 2
        assert expected_call_args == m_execute.call_args_list


@mock.patch("pycloudlib.instance.BaseInstance._do_restart")
@mock.patch("pycloudlib.instance.BaseInstance.get_boot_id")
@mock.patch("pycloudlib.instance.BaseInstance.wait")
@mock.patch("pycloudlib.instance.BaseInstance.wait_for_restart")
class TestRestart:
    """Test base restart behavior."""

    @pytest.fixture(autouse=True)
    def unchecked_mocks(self):
        """Mock things we don't want as test parameters."""
        with mock.patch("pycloudlib.instance.BaseInstance._sync_filesystem"):
            yield

    def test_no_wait(
        self,
        m_wait_for_restart,
        m_wait,
        m_boot_id,
        m_do_restart,
        concrete_instance_cls,
    ):
        """Test wait=False."""
        instance = concrete_instance_cls(key_pair=None)
        instance.restart(wait=False)
        assert m_do_restart.call_count == 1
        assert m_boot_id.call_count == 0
        assert m_wait_for_restart.call_count == 0
        assert m_wait.call_count == 0

    def test_instance_not_reachable(
        self,
        m_wait_for_restart,
        m_wait,
        m_boot_id,
        m_do_restart,
        concrete_instance_cls,
    ):
        """Test when instance is not reachable."""
        instance = concrete_instance_cls(key_pair=None)
        m_boot_id.side_effect = SSHException
        instance.restart(wait=True)
        assert m_do_restart.call_count == 1
        assert m_wait_for_restart.call_count == 0
        assert m_wait.call_count == 1

    def test_instance_reachable(
        self,
        m_wait_for_restart,
        m_wait,
        m_boot_id,
        m_do_restart,
        concrete_instance_cls,
    ):
        """Test when instance is reachable."""
        instance = concrete_instance_cls(key_pair=None)
        m_boot_id.side_effect = Result(
            "11111111-1111-1111-1111-111111111111", "", 0
        )
        instance.restart(wait=True)
        assert m_do_restart.call_count == 1
        assert m_wait_for_restart.call_count == 1
        assert m_wait.call_count == 0


class TestWaitForRestart:
    """Tests covering pycloudlib.instance.Instance.wait_for_restart."""

    @mock.patch.object(
        BaseInstance,
        "execute",
        side_effect=[
            Result("11111111-1111-1111-1111-111111111111", "", 0),
            Result("11111111-1111-1111-1111-111111111111", "", 0),
            Result("11111111-1111-1111-1111-111111111111", "", 0),
            Result("11111111-1111-1111-1111-111111111111", "", 0),
            Result("22222222-2222-2222-2222-222222222222", "", 0),
        ],
    )
    @mock.patch.object(BaseInstance, "_wait_for_cloudinit")
    @mock.patch("pycloudlib.instance.time.sleep")
    @mock.patch("pycloudlib.instance.time.time", return_value=1)
    def test_wait_for_restart(
        self, _m_time, _m_sleep, _m_wait_ci, m_execute, concrete_instance_cls
    ):
        """Test wait calls _wait_for_execute and waits till differing."""
        instance = concrete_instance_cls(key_pair=None)
        instance.wait_for_restart(
            old_boot_id="11111111-1111-1111-1111-111111111111"
        )
        assert m_execute.call_count == 5

    @pytest.mark.parametrize(
        "execute_side_effect",
        [
            SSHException,
            repeat(Result("11111111-1111-1111-1111-111111111111", "", 0)),
        ],
    )
    @mock.patch.object(BaseInstance, "execute")
    @mock.patch("pycloudlib.instance.time.sleep")
    @mock.patch("pycloudlib.instance.time.time")
    def test_boot_id_failure(
        self,
        m_time,
        m_sleep,
        m_execute,
        execute_side_effect,
        concrete_instance_cls,
    ):
        """Test wait calls when execute command fails."""
        m_execute.side_effect = execute_side_effect
        instance = concrete_instance_cls(key_pair=None)
        m_time.side_effect = [1, 2, 40 * 60, 40 * 60 + 1]
        expected_msg = (
            "Instance can't be reached after 40 minutes. "
            "Failed to obtain new boot id"
        )
        expected_call_args = [
            mock.call("cat /proc/sys/kernel/random/boot_id", no_log=True)
        ] * 2

        with pytest.raises(PycloudlibTimeoutError) as excinfo:
            instance.wait_for_restart(
                old_boot_id="11111111-1111-1111-1111-111111111111"
            )

        assert expected_msg == str(excinfo.value)
        assert m_sleep.call_count == 2
        assert expected_call_args == m_execute.call_args_list


# Disable this one because we're intentionally testing a protected member
# pylint: disable=protected-access
class TestWaitForCloudinit:
    """Tests covering pycloudlib.instance.Instance._wait_for_cloudinit."""

    def test_with_wait_available(self, concrete_instance_cls):
        """Test the happy path for instances with `status --wait`."""
        instance = concrete_instance_cls(key_pair=None)
        with mock.patch.object(instance, "execute") as m_execute:
            instance._wait_for_cloudinit()

        assert (
            mock.call(
                ["cloud-init", "status", "--wait", "--long"],
                description="waiting for start",
            )
            == m_execute.call_args
        )

    @mock.patch("time.sleep")
    def test_wait_on_target_not_active(self, _m_sleep, concrete_instance_cls):
        """Test that we wait for cloud-init is-active before calling status."""
        instance = concrete_instance_cls(key_pair=None)
        with mock.patch.object(
            instance,
            "execute",
            side_effect=[Result("", "", 0)] + [Result("", "", 1)] * 500,
        ) as m_execute:
            instance._wait_for_cloudinit()
        expected = [
            mock.call("command -v systemctl"),
            *(
                [
                    mock.call(
                        ["systemctl", "is-active", "cloud-init.target"],
                        no_log=True,
                    )
                ]
                * 300
            ),
            mock.call(
                ["cloud-init", "status", "--wait", "--long"],
                description="waiting for start",
            ),
        ]
        assert expected == m_execute.call_args_list



# This file is part of pycloudlib. See LICENSE file for license information.
"""Azure Cloud type."""
import base64
import logging
from typing import Dict, Optional

from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.network import NetworkManagementClient
from azure.mgmt.resource import ResourceManagementClient

from pycloudlib.azure import util
from pycloudlib.azure.instance import AzureInstance
from pycloudlib.cloud import BaseCloud, ImageType
from pycloudlib.config import ConfigFile
from pycloudlib.errors import (
    InstanceNotFoundError,
    NetworkNotFoundError,
    PycloudlibError,
)
from pycloudlib.util import get_timestamped_tag, update_nested

UBUNTU_DAILY_IMAGES = {
    "xenial": "Canonical:UbuntuServer:16.04-DAILY-LTS",
    "bionic": "Canonical:UbuntuServer:18.04-DAILY-LTS",
    "focal": "Canonical:0001-com-ubuntu-server-focal-daily:20_04-daily-lts",
    "impish": "Canonical:0001-com-ubuntu-server-impish-daily:21_10-daily",
    "jammy": "Canonical:0001-com-ubuntu-server-jammy-daily:22_04-daily-lts",
    "kinetic": "Canonical:0001-com-ubuntu-server-kinetic-daily:22_10-daily",
    "lunar": "Canonical:0001-com-ubuntu-server-lunar-daily:23_04-daily",
}

UBUNTU_DAILY_PRO_IMAGES = {
    "xenial": "Canonical:0001-com-ubuntu-pro-xenial:pro-16_04-lts",
    "bionic": "Canonical:0001-com-ubuntu-pro-bionic:pro-18_04-lts",
    "focal": "Canonical:0001-com-ubuntu-pro-focal:pro-20_04-lts",
    "jammy": "Canonical:0001-com-ubuntu-pro-jammy:pro-22_04-lts",
}

UBUNTU_DAILY_PRO_FIPS_IMAGES = {
    "xenial": "Canonical:0001-com-ubuntu-pro-xenial-fips:pro-fips-16_04-private",  # noqa
    "bionic": "Canonical:0001-com-ubuntu-pro-bionic-fips:pro-fips-18_04",
    "focal": "Canonical:0001-com-ubuntu-pro-focal-fips:pro-fips-20_04",
}

UBUNTU_RELEASE_IMAGES = {
    "xenial": "Canonical:UbuntuServer:16.04-LTS",
    "bionic": "Canonical:UbuntuServer:18.04-LTS",
    "focal": "Canonical:0001-com-ubuntu-server-focal:20_04-lts-gen2",
    "impish": "Canonical:0001-com-ubuntu-server-impish:21_10-gen2",
    "jammy": "Canonical:0001-com-ubuntu-server-jammy:22_04-lts-gen2",
}

logging.getLogger("azure.core.pipeline.policies.http_logging_policy").setLevel(
    logging.WARNING
)


class Azure(BaseCloud):
    """Azure Cloud Class."""

    _type = "azure"

    def __init__(
        self,
        tag: str,
        timestamp_suffix: bool = True,
        config_file: Optional[ConfigFile] = None,
        *,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        subscription_id: Optional[str] = None,
        tenant_id: Optional[str] = None,
        region: Optional[str] = None,
    ):
        """Initialize the connection to Azure.

        Azure will try to read user credentials form the /home/$USER/.azure
        folder. However, we can overwrite those credentials with the provided
        id parameters.

        Args:
            tag: string used to name and tag resources with
            timestamp_suffix: bool set True to append a timestamp suffix to the
                tag
            config_file: path to pycloudlib configuration file
            client_id: user's client id
            client_secret: user's client secret access key
            subscription_id: user's subscription id key
            tenant_id: user's tenant id key
            region: The region where the instance will be created
        """
        super().__init__(
            tag,
            timestamp_suffix,
            config_file,
            required_values=[
                client_id,
                client_secret,
                subscription_id,
                tenant_id,
            ],
        )

        self._log.debug("logging into Azure")
        self.location = region or self.config.get("region") or "centralus"
        self.username = "ubuntu"

        self.registered_instances: Dict[str, AzureInstance] = {}
        self.registered_images: Dict[str, dict] = {}

        config_dict = {}

        client_id = client_id or self.config.get("client_id")
        if client_id:
            config_dict["clientId"] = client_id

        client_secret = client_secret or self.config.get("client_secret")
        if client_secret:
            config_dict["clientSecret"] = client_secret

        subscription_id = subscription_id or self.config.get("subscription_id")
        if subscription_id:
            config_dict["subscriptionId"] = subscription_id

        tenant_id = tenant_id or self.config.get("tenant_id")
        if tenant_id:
            config_dict["tenantId"] = tenant_id

        self.resource_client = util.get_client(
            ResourceManagementClient, config_dict
        )

        self.network_client = util.get_client(
            NetworkManagementClient, config_dict
        )

        self.compute_client = util.get_client(
            ComputeManagementClient, config_dict
        )

        self.resource_group = self._create_resource_group()
        self.base_tag = tag

    def image_serial(self, image_id):
        """Find the image serial of the latest daily image for a particular release.

        Args:
            image_id: string, Ubuntu image id

        Returns:
            string, serial of latest image

        """
        raise NotImplementedError

    def _create_network_security_group(self, inbound_ports):
        """Create a network security group.

        This method creates a network security groups that allows the user
        to ssh into the machine and execute commands.

        Args:
            inbound_ports: List of strings, optional inbound ports
                           to enable in the instance.

        Returns:
            The network security object created by Azure

        """
        if not inbound_ports:
            inbound_ports = []

        # We need to guarantee that the 22 port is enabled
        # here, otherwise we will not be able to ssh into it
        if "22" not in inbound_ports:
            inbound_ports = ["22"] + inbound_ports

        security_group_name = "{}-sgn".format(self.tag)
        nsg_group = self.network_client.network_security_groups

        self._log.debug("Creating Azure network security group")

        security_rules = []

        # The lower the number, the higher is the priority of the rule.
        # We are assuming here that the SSH rule will be the first item
        # in the list
        priority = 300
        for port in inbound_ports:
            security_rules.append(
                {
                    "name": "port-{}".format(port),
                    "priority": priority,
                    "protocol": "TCP",
                    "access": "Allow",
                    "direction": "Inbound",
                    "sourceAddressPrefix": "*",
                    "sourcePortRange": "*",
                    "destinationAddressPrefix": "*",
                    "destinationPortRange": port,
                }
            )
            priority += 10

        nsg_poller = nsg_group.begin_create_or_update(
            resource_group_name=self.resource_group.name,
            network_security_group_name=security_group_name,
            parameters={
                "location": self.location,
                "security_rules": security_rules,
            },
        )

        return nsg_poller.result()

    def _create_resource_group(self):
        """Create a resource group.

        This method creates an Azure resource group. Every other component that
        we create will be contained into this resource group. This means that
        if we delete this resource group, we delete all resources associated
        with it.

        Returns:
            The resource group created by Azure

        """
        resource_name = "{}-rg".format(self.tag)
        self._log.debug("Creating Azure resource group")

        return self.resource_client.resource_groups.create_or_update(
            resource_name,
            {"location": self.location, "tags": {"name": self.tag}},
        )

    def _create_virtual_network(self, address_prefixes=None):
        """Create a virtual network.

        This method creates an Azure virtual network to be used
        when provisioning a subnet.

        Args:
            address_prefixes:  list of strings, A list of address prefixes
                               to be used in this virtual network.
        Returns:
            The virtual network created by Azure

        """
        if address_prefixes is None:
            address_prefixes = ["10.0.0.0/16"]

        virtual_network_name = "{}-vnet".format(self.tag)

        self._log.debug("Creating Azure virtual network")
        network_poller = (
            self.network_client.virtual_networks.begin_create_or_update(
                self.resource_group.name,
                virtual_network_name,
                {
                    "location": self.location,
                    "address_space": {"address_prefixes": address_prefixes},
                    "tags": {"name": self.tag},
                },
            )
        )

        return network_poller.result()

    def _create_subnet(self, vnet_name, address_prefix="10.0.0.0/24"):
        """Create a subnet.

        This method creates an Azure subnet to be used when
        provisioning a network interface.

        Args:
            address_prefix: string, An address prefix to be used for
                            this subnet.

        Returns:
            The subnet created by Azure

        """
        subnet_name = "{}-subnet".format(self.tag)

        self._log.debug("Creating Azure subnet")
        subnet_poller = self.network_client.subnets.begin_create_or_update(
            self.resource_group.name,
            vnet_name,
            subnet_name,
            {"address_prefix": address_prefix, "tags": {"name": self.tag}},
        )

        return subnet_poller.result()

    def _create_ip_address(self):
        """Create an ip address.

        This method creates an Azure ip address to be used when
        provisioning a network interface

        Returns:
            The ip address created by Azure

        """
        ip_name = "{}-ip".format(self.tag)

        self._log.debug("Creating Azure ip address")
        ip_poller = (
            self.network_client.public_ip_addresses.begin_create_or_update(
                self.resource_group.name,
                ip_name,
                {
                    "location": self.location,
                    "sku": {"name": "Standard"},
                    "public_ip_allocation_method": "Static",
                    "rpublic_ip_address_version": "IPV4",
                    "tags": {"name": self.tag},
                },
            )
        )

        return ip_poller.result()

    def _create_network_interface_client(
        self, ip_address_id, subnet_id, nsg_id
    ):
        """Create a network interface client.

        This method creates an Azure network interface to be used when
        provisioning a virtual machine

        Args:
            ip_address_id: string, The ip address id
            subnet_id: string, the subnet id
            nsg_id: string, the network security group id

        Returns:
            The ip address created by Azure

        """
        nic_name = "{}-nic".format(self.tag)
        ip_config_name = "{}-ip-config".format(self.tag)

        self._log.debug("Creating Azure network interface")
        nic_poller = (
            self.network_client.network_interfaces.begin_create_or_update(
                self.resource_group.name,
                nic_name,
                {
                    "location": self.location,
                    "ip_configurations": [
                        {
                            "name": ip_config_name,
                            "subnet": {"id": subnet_id},
                            "public_ip_address": {"id": ip_address_id},
                        }
                    ],
                    "network_security_group": {"id": nsg_id},
                    "tags": {"name": self.tag},
                },
            )
        )

        return nic_poller.result()

    def _create_vm_parameters(
        self, name, image_id, instance_type, nic_id, user_data
    ):
        """Create the virtual machine parameters to be used for provision.

        Composes the dict that will be used to provision an Azure virtual
        machine. We check if the user has passed user_data and the type of
        image_id we are receiving, which can be snapshots ids or not.

        Args:
            name: string, The name of the virtual machine.
            image_id: string, The identifier of an image.
            instance_type: string, Type of instance to create.
            nic_id: string, The network interface id.
            user_data: string, The user data to be passed to the
                       virtual machine.

        Returns:
            A dict containing the parameters to provision a virtual machine.

        """
        vm_parameters = {
            "location": self.location,
            "hardware_profile": {"vm_size": instance_type},
            "storage_profile": {"image_reference": {}},
            "os_profile": {
                "computer_name": name,
                "admin_username": self.username,
                "linux_configuration": {
                    "ssh": {
                        "public_keys": [
                            {
                                "path": "/home/{}/.ssh/authorized_keys".format(
                                    self.username
                                ),
                                "key_data": self.key_pair.public_key_content,
                            }
                        ]
                    },
                    "disable_password_authentication": True,
                },
            },
            "network_profile": {
                "network_interfaces": [
                    {
                        "id": nic_id,
                    }
                ]
            },
            "tags": {"name": self.tag},
        }

        if user_data:
            # We need to encode the user_data into base64 before sending
            # it to the virtual machine.
            vm_parameters["os_profile"]["custom_data"] = base64.b64encode(
                user_data.encode()
            ).decode()

        vm_parameters["storage_profile"][
            "image_reference"
        ] = util.get_image_reference_params(image_id)

        # We can have pro images from two different sources; marketplaces
        # and snapshots. A snapshot image does not have the necessary metadata
        # encoded in the image_id to create the 'plan' dict. In this case,
        # we get the necessary info from the registered_images dict
        # where we store the required metadata about any snapshot created by
        # pycloudlib.
        registered_image = self.registered_images.get(image_id)
        if util.is_pro_image(image_id, registered_image):
            vm_parameters["plan"] = util.get_plan_params(
                image_id, registered_image
            )

        return vm_parameters

    def _create_virtual_machine(
        self, image_id, instance_type, nic_id, user_data, name, vm_params=None
    ):
        """Create a virtual machine.

        This method provisions an Azure virtual machine for the image_id
        provided by the user.

        Args:
            image_id: string, The image to be used when provisiong
                      a virtual machine.
            instance_type: string, Type of instance to create
            nic_id: string, The network interface to used for this
                    virtual machine.
            user_data: string, user data used by cloud-init when
                       booting the virtual machine.
            name: string, optional name to provide when creating the vm.
            vm_params: dict containing values as vm_params to send to
                    virtual_machines.begin_create_or_update.

        Returns:
            The virtual machine created by Azure

        """
        if not name:
            name = "{}-vm".format(self.tag)
        params = self._create_vm_parameters(
            name, image_id, instance_type, nic_id, user_data
        )
        if vm_params:
            update_nested(params, vm_params)
        self._log.debug("Creating Azure virtual machine: %s", name)
        vm_poller = (
            self.compute_client.virtual_machines.begin_create_or_update(
                self.resource_group.name,
                name,
                params,
            )
        )

        return vm_poller.result()

    def delete_image(self, image_id, **kwargs):
        """Delete an image from Azure.

        Args:
            image_id: string, The id of the image to be deleted
        """
        image_name = util.get_resource_name_from_id(image_id)
        resource_group_name = util.get_resource_group_name_from_id(image_id)

        delete_poller = self.compute_client.images.begin_delete(
            resource_group_name=resource_group_name, image_name=image_name
        )

        delete_poller.wait()

        if delete_poller.status() == "Succeeded":
            self._log.debug("Image %s was deleted", image_id)
            del self.registered_images[image_id]
        else:
            self._log.debug(
                "Error deleting %s. Status: %d",
                image_id,
                delete_poller.status(),
            )

    def _get_image(self, release, image_map):
        image_id = image_map.get(release)
        if image_id is None:
            msg = "No Ubuntu image found for {}. Expected one of: {}"
            raise ValueError(msg.format(release, " ".join(image_map.keys())))

        return image_id

    def released_image(self, release):
        """Get the released image.

        Args:
            release: string, Ubuntu release to look for
        Returns:
            string, id of latest image

        """
        self._log.debug("finding release Ubuntu image for %s", release)
        return self._get_image(release, UBUNTU_RELEASE_IMAGES)

    def _get_images_dict(self, image_type: ImageType):
        if image_type == ImageType.GENERIC:
            return UBUNTU_DAILY_IMAGES

        if image_type == ImageType.PRO:
            return UBUNTU_DAILY_PRO_IMAGES

        if image_type == ImageType.PRO_FIPS:
            return UBUNTU_DAILY_PRO_FIPS_IMAGES

        raise ValueError("Invalid image_type")

    def daily_image(
        self,
        release: str,
        *,
        image_type: ImageType = ImageType.GENERIC,
        **kwargs,
    ):
        """Find the image info for the latest daily image for a given release.

        Args:
            release: string, Ubuntu release to look for.

        Returns:
            A string representing an Ubuntu image

        """
        self._log.debug("finding daily Ubuntu image for %s", release)
        return self._get_image(release, self._get_images_dict(image_type))

    def _check_for_network_interfaces(self):
        """
        Check for existing networking interfaces in instance resource group.

        Check if we already have a network interface that is not attached to
        any virtual machines in the instance resource group. If we have one
        of those reoources, we just return it.

        Returns:
            An Azure network interface resource

        """
        all_nics = self.network_client.network_interfaces.list(
            resource_group_name=self.resource_group.name
        )

        for nic in all_nics:
            if nic.virtual_machine is None:
                return nic

        return None

    def launch(
        self,
        image_id,
        instance_type="Standard_DS1_v2",
        user_data=None,
        wait=True,
        name=None,
        inbound_ports=None,
        **kwargs,
    ):
        """Launch virtual machine on Azure.

        Args:
            image_id: string, Ubuntu image to use
            user_data: string, user-data to pass to virtual machine
            wait: boolean, wait for instance to come up
            name: string, optional name to give the vm when launching.
                  Default results in a name of <tag>-vm
            inbound_ports: List of strings, optional inbound ports
                           to enable in the instance.
            kwargs: dict, other named arguments to provide to
                    virtual_machines.begin_create_or_update

        Returns:
            Azure Instance object
        Raises: ValueError on invalid image_id
        """
        # pylint: disable-msg=too-many-locals
        if not image_id:
            raise ValueError(
                f"{self._type} launch requires image_id param."
                f" Found: {image_id}"
            )
        self._log.debug("Launching Azure virtual machine: %s", image_id)

        # For every new launch, we need to update the tag, since
        # we are using it as a base for the name of all the
        # resources we are creating.
        self.tag = get_timestamped_tag(self.base_tag)

        if self.resource_group is None:
            self.resource_group = self._create_resource_group()

        # We will not reuse existing network interfaces if we need to customize
        # it to enable more ports. The rationale for is that we want to reuse
        # those resources only if they are generic enough
        nic = None
        if not inbound_ports:
            # Check if we already have an existing network interface that is
            # not attached to a virtual machine. If we have, we will just
            # use it
            nic = self._check_for_network_interfaces()

        if nic is None:
            self._log.debug(
                "Could not find a network interface. Creating one now"
            )
            virtual_network = self._create_virtual_network()
            self._log.debug(
                "Created virtual network with name: %s", virtual_network.name
            )

            subnet = self._create_subnet(vnet_name=virtual_network.name)
            self._log.debug("Created subnet with name: %s", subnet.name)

            ip_address = self._create_ip_address()
            ip_address_str = ip_address.ip_address
            self._log.debug(
                "Created ip address with name: %s", ip_address.name
            )

            network_security_group = self._create_network_security_group(
                inbound_ports=inbound_ports
            )
            self._log.debug(
                "Created network security group with name: %s",
                network_security_group.name,
            )

            nic = self._create_network_interface_client(
                ip_address_id=ip_address.id,
                subnet_id=subnet.id,
                nsg_id=network_security_group.id,
            )

            self._log.debug(
                "Created network interface with name: %s", nic.name
            )
        else:
            ip_address_str = self._retrieve_ip_from_network_interface(nic=nic)
            self._log.debug(
                "Found network interface: %s. Reusing it", nic.name
            )

        vm = self._create_virtual_machine(
            image_id=image_id,
            instance_type=instance_type,
            nic_id=nic.id,
            user_data=user_data,
            name=name,
            vm_params=kwargs.get("vm_params", None),
        )

        instance_info = {
            "vm": vm,
            "ip_address": ip_address_str,
            "rg_name": self.resource_group.name,
        }

        instance = AzureInstance(
            key_pair=self.key_pair,
            client=self.compute_client,
            instance=instance_info,
        )

        if wait:
            instance.wait()

        self.registered_instances[vm.name] = instance
        return instance

    def _create_ssh_resource(self, key_name):
        """Create a ssh resource.

        This method creates an Azure ssh resource to be associated
        with a resource group.

        Args:
            key_name: string, The name of the ssh resource.

        """
        self.compute_client.ssh_public_keys.create(
            self.resource_group.name,
            key_name,
            parameters={"location": self.location, "tags": {"name": self.tag}},
        )

    def create_key_pair(self, key_name):
        """Create a pair of ssh keys.

        This method creates an a pair of ssh keys in
        the class resource group.

        Args:
            key_name: string, The name of the ssh resource.

        """
        self._create_ssh_resource(key_name)

        ssh_call = self.compute_client.ssh_public_keys.generate_key_pair(
            resource_group_name=self.resource_group.name,
            ssh_public_key_name=key_name,
        )

        # Azure's SDK returns multi-line DOS format for pubkeys.
        # OpenSSH doesn't like this format and ignores it resulting in
        # Unauthorized key errors. Issue: #88
        return ssh_call.public_key.replace("\r\n", ""), ssh_call.private_key

    def list_keys(self):
        """List all ssh keys in the class resource group."""
        ssh_public_keys = self.compute_client.ssh_public_keys

        return [
            ssh.name
            for ssh in ssh_public_keys.list_by_resource_group(
                self.resource_group.name
            )
        ]

    def delete_key(self, key_name):
        """Delete a ssh key from the class resource group.

        Args:
            key_name: string, The name of the ssh resource.

        """
        ssh_public_keys = self.compute_client.ssh_public_keys
        ssh_public_keys.delete(
            resource_group_name=self.resource_group.name,
            ssh_public_key_name=key_name,
        )

    def use_key(self, public_key_path, private_key_path=None, name=None):
        """Use an existing already uploaded key.

        Args:
            public_key_path: path to the public key to upload
            private_key_path: path to the private key to upload
            name: name to reference key by

        """
        if not name:
            name = self.tag
        super().use_key(public_key_path, private_key_path, name)

    def _get_instances(self):
        """Return an iterable of Azure instances related to a subscription id.

        Returns:
            An list of azure virtual machine associated with the subscription
            id

        """
        return self.compute_client.virtual_machines.list_all()

    def _retrieve_ip_from_network_interface(self, nic):
        """Retrieve the ip address associated with a network interface.

        Args:
            nic: An Azure network interface resource

        Return:
            A string representing the network interface ip address

        """
        ip_address_id = nic.ip_configurations[0].public_ip_address.id
        all_ips = self.network_client.public_ip_addresses.list_all()

        for ip_address in all_ips:
            if ip_address.id == ip_address_id:
                return ip_address.ip_address

        raise PycloudlibError(
            f"""
            Error locationg the ip address: {ip_address_id}.
            This ip address was not found in this subscription.
            """
        )

    def _retrive_instance_ip(self, instance):
        """Retrieve public ip address of instance.

        Args:
            instance: An Azure Virtual Machine object

        Returns:
            A string represeting the instance ip_address

        """
        # Right now, we are only supporting getting the ip address for
        # virtual machines with only one network profile attached to it
        nic_id = instance.network_profile.network_interfaces[0].id
        all_nics = self.network_client.network_interfaces.list_all()

        instance_nic = None
        for nic in all_nics:
            if nic.id == nic_id:
                instance_nic = nic

        if instance_nic is None:
            raise NetworkNotFoundError(resource_id=nic_id)

        return self._retrieve_ip_from_network_interface(nic=instance_nic)

    def get_instance(self, instance_id, search_all=False):
        """Get an instance by id.

        Args:
            instance_id: string, The instance name to search by
            search_all: boolean, Flag that indicates that if we should search
                        for the instance in the entire reach of the
                        subsctription id. If false, we will search only
                        in the resource group created by this instance.

        Returns:
            An instance object to use to manipulate the instance further.

        """
        if search_all:
            all_instances = self._get_instances()

            for instance in all_instances:
                if instance.name == instance_id:
                    ip_address = self._retrive_instance_ip(instance)
                    resource_group_name = util.get_resource_group_name_from_id(
                        instance.id
                    )

                    instance_info = {
                        "vm": instance,
                        "ip_address": ip_address,
                        "rg_name": resource_group_name,
                    }
                    azure_instance = AzureInstance(
                        key_pair=self.key_pair,
                        client=self.compute_client,
                        instance=instance_info,
                    )

                    self.registered_instances[instance.name] = azure_instance
                    return azure_instance

            raise InstanceNotFoundError(instance_id)

        if instance_id in self.registered_instances:
            instance = self.registered_instances[instance_id]

            if instance.status == "deleted":
                raise PycloudlibError(
                    f"The image {instance_id} was already deleted"
                )

            return instance

        raise InstanceNotFoundError(resource_id=instance_id)

    def snapshot(
        self, instance, clean=True, delete_provisioned_user=True, **kwargs
    ):
        """Snapshot an instance and generate an image from it.

        Args:
            instance: Instance to snapshot
            clean: Run instance clean method before taking snapshot
            delete_provisioned_user: Deletes the last provisioned user
            kwargs: Other named arguments specific to this implementation

        Returns:
            An image id string

        """
        if clean:
            instance.clean()
        user = "+user" if delete_provisioned_user else ""
        instance.execute("sudo waagent -deprovision{} -force".format(user))
        instance.shutdown(wait=True)
        instance.generalize()

        self._log.debug("creating custom image from instance %s", instance.id)

        image_poller = self.compute_client.images.begin_create_or_update(
            resource_group_name=self.resource_group.name,
            image_name="%s-%s" % (self.tag, "image"),
            parameters={
                "location": self.location,
                "source_virtual_machine": {"id": instance.id},
                "tags": {"name": self.tag, "src-image-id": instance.image_id},
            },
        )

        image = image_poller.result()

        image_id = image.id
        image_name = image.name

        self.registered_images[image_id] = {
            "name": image_name,
            "sku": instance.sku,
            "offer": instance.offer,
        }

        return image_id

    def delete_resource_group(self):
        """Delete a resource group."""
        if self.resource_group:
            self.resource_client.resource_groups.begin_delete(
                resource_group_name=self.resource_group.name
            )

            self.resource_group = None
